from __future__ import annotations

import json
import concurrent.futures
from gcp_pal.utils import try_import

from gcp_pal.schema import enforce_schema
from gcp_pal.utils import (
    is_dataframe,
    log,
    ClientHandler,
    ModuleHandler,
    get_default_arg,
)


class Firestore:
    """
    Class for operating Firestore
    """

    def __init__(self, path=None, project=None):
        """
        Args:
        - path (str): Path to the Firestore document or collection
        - project (str): Project ID

        Storage-equivalent examples:
        - `Firestore().ls()` -> List all collections
        - `Firestore("collection").ls()` -> List all documents in collection "collection"
        - `Firestore("bucket/path").read()` -> Read from Firestore "bucket/path"
        - `Firestore("gs://project/bucket/path").read()` -> Read from Firestore "bucket/path"
        - `Firestore("gs://project/bucket/path").write(data)` -> Write to Firestore "bucket/path"
        - `Firestore("gs://project/bucket/path").delete()` -> Delete from Firestore "bucket/path"
        """
        self.project = project or get_default_arg("project")
        self.path = path

        # Only initialize the client once per project
        self.firestore = ModuleHandler("google.cloud").please_import(
            "firestore", who_is_calling="Firestore"
        )
        self.client = ClientHandler(self.firestore.Client).get(project=self.project)

    def __repr__(self):
        return f"Firestore({self.path})"

    def _parse_path(self, method="get"):
        """
        Parse the path into project, bucket, and path. This allows Firestore to be used in the same way as GCS.

        Returns:
        - Path elements (list): List of alternating path elements [collection, document, collection, ...]

        Examples:
        - Firestore("collection/document")._parse_path() -> ["collection", "document"]
        - Firestore("gs://project/bucket/collection/document")._parse_path() -> ["collection", "document"]
        - Firestore("gs://project/bucket/output/data.csv")._parse_path() -> ["output", "data.csv"]
        """
        if self.path is None:
            return None
        if "gs://" in self.path:
            # gs://project/bucket/path -> project, bucket, path
            # path -> collection/document/../collection/document
            self.path = self.path.replace("gs://", "")
            self.bucket, self.path = self.path.split("/", 1)
        path_elements = self.path.split("/")
        return path_elements

    def get(self, method=None):
        """
        Get a reference to a Firestore document or collection.

        Returns:
        - Firestore reference (DocumentReference or CollectionReference)

        Examples:
        - Firestore("collection/document").get() -> Reference to document "collection/document"
        - Firestore("collection").get() -> Reference to collection "collection"
        """
        path_elements = self._parse_path(method)
        if path_elements is None:
            return None
        doc_ref = self.client.collection(path_elements.pop(0))
        ref_type = "document"
        while len(path_elements) > 0:
            doc_ref = getattr(doc_ref, ref_type)(path_elements.pop(0))
            ref_type = "document" if ref_type == "collection" else "collection"
        return doc_ref

    def async_read(self, paths_list, allow_empty=False, apply_schema=False, schema={}):
        """
        Read a list of paths from Firestore asynchronously.

        Args:
        - paths_list (list): List of paths to read from Firestore
        - allow_empty (bool): If True, return an empty DataFrame if the document is empty
        - apply_schema (bool): If True, apply the schema from FIRESTORE_SCHEMAS.
                               Also converts the output to a DataFrame.
        """
        with concurrent.futures.ThreadPoolExecutor(max_workers=50) as executor:
            futures = [
                executor.submit(
                    Firestore(path).read,
                    allow_empty=allow_empty,
                    apply_schema=apply_schema,
                    schema=schema,
                )
                for path in paths_list
            ]
            concurrent.futures.wait(futures)
            output = {
                path.split("/")[-1]: future.result()
                for path, future in zip(paths_list, futures)
            }
        return output

    def read(self, allow_empty=False, apply_schema=False, schema={}, path_schemas={}):
        """
        Read from Firestore

        Args:
        - allow_empty (bool): If True, return an empty DataFrame if the document is empty
        - apply_schema (bool): If True, apply the schema from FIRESTORE_SCHEMAS.
                               Also converts the output to a DataFrame.
        - schema (dict): Schema to enforce on the output
        - path_schemas (dict): Schemas to enforce on specific paths.

        Returns:
        - Output from Firestore (DataFrame or dict)

        Examples:
        - Firestore("coll/doc").read() -> Read from Firestore "coll/doc"
        - Firestore("coll").read() -> Read all docs from Firestore collection "coll"
        - Firestore("coll/doc").read(apply_schema=True, path_schemas={"coll/doc": {"a": int}})
                                   -> Read from Firestore "coll/doc" and enforce schema on "a"
        """
        if path_schemas:
            path_schemas = path_schemas.get(self.path, {})
            schema = {**schema, **path_schemas}
        doc_ref = self.get(method="get")
        if self._ref_type(doc_ref) == "collection":
            # If the reference is a collection, return a list of documents
            paths_list = [f"{self.path}/{doc.id}" for doc in doc_ref.stream()]
            return self.async_read(
                paths_list,
                allow_empty=allow_empty,
                apply_schema=apply_schema,
                schema=schema,
            )
        output = doc_ref.get().to_dict()
        metadata = {}
        object_type = None
        dtypes = None
        if output is None and allow_empty:
            output = {}
        if set(output.keys()) in [{"data"}, {"data", "metadata"}]:
            metadata = output.get("metadata", {})
            object_type = metadata.get("object_type", None)
            dtypes = metadata.get("dtypes", None)
            output = output.get("data", output)
            if dtypes is not None:
                apply_schema = True
        if apply_schema:
            if object_type == "<class 'pandas.core.frame.DataFrame'>":
                try_import("pandas", "DataFrame")
                from pandas import DataFrame

                output = DataFrame(output)
                output = output.reset_index(drop=True)
            output = enforce_schema(output, schema=schema, dtypes=dtypes)
        log(f"Firestore - read {self.path}")
        return output

    def write(self, data, columns=None):
        """
        Write to Firestore

        Args:
        - data (DataFrame or dict): Data to write to Firestore
        - columns (list): Columns to write from the DataFrame

        Returns:
        - True if successful

        Examples:
        - Firestore("coll/doc").write(data) -> Write data to Firestore "coll/doc"
        """
        doc_ref = self.get(method="set")
        dtypes, object_type = None, None
        object_type = str(type(data))
        if is_dataframe(data):
            if columns is not None:
                data = data[columns]
            data.reset_index(drop=True, inplace=True)
            dtypes = data.dtypes.astype(str).to_dict()
            data = data.to_json()
        try:
            doc_ref.set(data)
        except ValueError as e:
            # This happens if the data is a dictionary with integer keys
            doc_ref.set(json.loads(json.dumps(data)))
        except AttributeError as e:
            # This happens if the data is a list of dictionaries
            if isinstance(data, str):
                # This happens if the data came from DataFrame.to_json()
                data = json.loads(data)
            output = {"data": data, "metadata": {}}
            if dtypes is not None:
                output["metadata"]["dtypes"] = dtypes
            if object_type is not None:
                output["metadata"]["object_type"] = object_type
            doc_ref.set(output)
        log(f"Firestore - written {self.path}")
        return True

    def create(self, **kwargs):
        """
        Create an empty Firestore document or collection.
        """
        ref = self.get()
        ref_type = self._ref_type(ref)
        if ref_type == "document":
            ref.set({})
        elif ref_type == "collection":
            ref.add({})
        else:
            raise ValueError("Unsupported Firestore reference type.")
        log(f"Firestore - created {ref_type}: Firestore/{self.path}")
        return True

    def delete(self):
        """
        Recursively deletes documents and collections from Firestore.
        """
        ref = self.get()
        ref_type = self._ref_type(ref)
        if ref_type == "document":
            self._delete_document(ref)
        elif ref_type == "collection":
            self._delete_collection(ref)
        else:
            raise ValueError("Unsupported Firestore reference type.")
        log(f"Firestore - deleted {ref_type}: Firestore/{self.path}")
        return True

    def _delete_document(self, doc_ref):
        """
        Deletes a document and all of its collections.
        """
        collections = doc_ref.collections()
        for collection in collections:
            self._delete_collection(collection)
        doc_ref.delete()

    def _delete_collection(self, col_ref):
        """
        Deletes a collection and all of its documents.
        """
        docs = col_ref.stream()
        with concurrent.futures.ThreadPoolExecutor(max_workers=50) as executor:
            futures = [
                executor.submit(self._delete_document, doc.reference) for doc in docs
            ]
            concurrent.futures.wait(futures)

    def _ref_type(self, doc_ref):
        is_doc_ref = isinstance(doc_ref, self.firestore.DocumentReference)
        is_coll_ref = isinstance(doc_ref, self.firestore.CollectionReference)
        return "document" if is_doc_ref else "collection" if is_coll_ref else None

    def ls(self, path=None) -> list[str]:
        """
        List all documents in a collection or all collections in a document.

        Args:
        - path (str): Path to the Firestore document or collection from base path

        Returns:
        - List of documents or collections
        """
        if path is not None:
            self.path = self.path + "/" + path
        ref = self.get()
        if ref is None:
            output = [col.id for col in self.client.collections()]
            log(f"Firestore - collections listed.")
            return output
        ref_type = self._ref_type(ref)
        if ref_type == "document":
            output = [doc.id for doc in ref.collections()]
        elif ref_type == "collection":
            output = [doc.id for doc in ref.stream()]
        else:
            raise ValueError("Unsupported Firestore reference type.")
        log(f"Firestore - collections listed.")
        return output

    def exists(self):
        """
        Check if a document or collection exists in Firestore.
        """
        ref = self.get()
        ref_type = self._ref_type(ref)
        if ref_type == "document":
            exists = ref.get().exists
        elif ref_type == "collection":
            exists = bool(list(ref.stream()))
        else:
            raise ValueError("Unsupported Firestore reference type.")
        return exists


if __name__ == "__main__":
    data = {
        "a": [1, 2, 3],
        "b": ["a", "b", "c"],
        "c": [1, 2, 3],
    }
    import pandas as pd

    data = pd.DataFrame(data)
    collection_name = "test_collection"
    Firestore(f"{collection_name}/test_document1").write(data)
    Firestore(f"{collection_name}/test_document2").write(data)
    Firestore(f"{collection_name}/test_document3").write(data)
    print(Firestore(collection_name).ls())
    print(Firestore().ls())
    output = Firestore(collection_name).read(apply_schema=True)
    print(output)
    Firestore(collection_name).delete()



from gcp_pal.utils import log, ModuleHandler, get_default_arg


class Request:
    """
    Class for making authorized requests to a cloud run service.

    Examples:
    - `Request("https://[CLOUD_RUN_URL]").get()` -> Get request to cloud run service
    - `Request("https://[CLOUD_RUN_URL]").post({"key": "value"})` -> Post request to cloud run service
    """

    _identity_token = None

    def __init__(self, url, project=None, service_account=None):
        """
        Args:
        - url (str): URL of the cloud run service
        - project (str): Project ID
        """
        self.url = url

        self.requests = ModuleHandler("requests").please_import(
            who_is_calling="Request"
        )
        self.google_auth = ModuleHandler("google.auth").please_import(
            who_is_calling="Request"
        )
        self.id_token = ModuleHandler("google.oauth2").please_import(
            "id_token", who_is_calling="Request"
        )
        self.AuthRequest = ModuleHandler(
            "google.auth.transport.requests"
        ).please_import("Request", who_is_calling="Request")
        self.exceptions = ModuleHandler("google.auth.exceptions").please_import(
            who_is_calling="Request"
        )

        scopes = ["https://www.googleapis.com/auth/cloud-platform"]
        self.credentials, self.project = self.google_auth.default(scopes=scopes)

        self.project = project or self.project or get_default_arg("project")

        default_sa = f"{self.project}@{self.project}.iam.gserviceaccount.com"
        self.service_account = service_account or default_sa

        self.identity_token = self.get_identity_token()
        self.headers = {
            "Authorization": f"Bearer {self.identity_token}",
            "Content-type": "application/json",
        }
        self.args = {}

    def __repr__(self):
        return f"Request({self.url})"

    def get_identity_token(self):
        # Attempt to fetch an identity token for the given URL
        auth_req = self.AuthRequest()
        try:
            # Ensure the credentials are valid and refreshed
            if not self.credentials.valid:
                self.credentials.refresh(auth_req)

            # The audience URL should be the URL of the cloud function or service you are accessing.
            # Make sure this matches exactly what's expected by the service.
            token = self.id_token.fetch_id_token(auth_req, self.url)
            return token
        except self.exceptions.DefaultCredentialsError:
            log(f"Request - Fetching credentials via access token.")
            return self._fetch_identity_access_token(auth_req)
        except self.exceptions.RefreshError as e:
            log(f"Request - Error refreshing credentials: {e}")
            return None

    def _fetch_identity_access_token(self, auth_request):
        """
        Fetches an access token from the identity token.

        Args:
        - auth_request (google.auth.transport.requests.Request): Request object

        Returns:
        - str: Access token
        """
        self.credentials.refresh(auth_request)
        access_token = self.credentials.token
        iam_url = "https://iamcredentials.googleapis.com"
        url = f"{iam_url}/v1/projects/-/serviceAccounts/{self.service_account}:generateIdToken"
        params = {"audience": self.url}
        headers = {
            "Content-type": "text/json; charset=utf-8",
            "Metadata-Flavor": "Google",
            "Authorization": f"Bearer {access_token}",
        }
        response = self.requests.post(url, params=params, headers=headers)
        if response.status_code == 200:
            response_json = response.json()
            return response_json.get("token")
        else:
            if "iam.serviceAccounts.getOpenIdToken" in response.text:
                role = "Service Account Token Creator"
                acc = self.service_account
                log(
                    f"Request - Error: Ensure the service account {acc} has the '{role}' role."
                )
            else:
                log(f"Request - Error fetching identity token: {response.text}")
        return None

    def post(self, payload=None, **kwargs):
        arg_name = "data" if isinstance(payload, dict) else "json"
        self.args = {arg_name: payload, "headers": self.headers, **kwargs}
        response = self.requests.post(self.url, **self.args)
        return response

    def get(self, **kwargs):
        response = self.requests.get(self.url, headers=self.headers, **kwargs)
        return response

    def put(self, payload=None, **kwargs):
        arg_name = "data" if isinstance(payload, dict) else "json"
        self.args = {arg_name: payload, "headers": self.headers, **kwargs}
        response = self.requests.put(self.url, **self.args)
        return response


if __name__ == "__main__":
    uri = "https://python-roh-jfzraqzsma-nw.a.run.app"
    payload = {"task_name": "events"}
    response = Request(uri).post(payload)
    print(response.json())
    print(response.status_code)



import importlib
from gcp_pal.utils import try_import


class ModuleHandler:
    """
    Class to handle module import and instantiation.
    """

    _imported = {}

    def __init__(self, module_name: str = None):
        """
        Initialize the ModuleHandler.

        Args:
        - module_name: The name of the module.
        """
        self.module_name = module_name

    def please_import(
        self,
        thing: str = None,
        who_is_calling: str = None,
        errors: str = "raise",
    ):
        """
        Import the module.

        Args:
        - thing: The name of the thing to import (if any). If None, import the module.
        - who_is_calling: The name of the function/class that is calling this method (for information only).
        """
        full_name = self.module_name
        if thing is not None:
            full_name += f".{thing}"

        if full_name in ModuleHandler._imported:
            # Load the module from the cache:
            return ModuleHandler._imported[full_name]

        # Try to import the module. This will show the missing dependencies.
        try_import(full_name, who_is_calling, errors)

        # If no errors, we can import the module (this is fast):
        try:
            module = importlib.import_module(full_name)
        except ModuleNotFoundError:
            module = importlib.import_module(self.module_name)
            if thing is not None:
                module = getattr(module, thing)

        # Save the imported module to the cache:
        ModuleHandler._imported[full_name] = module

        return module


if __name__ == "__main__":
    bq_client1 = ModuleHandler("google.cloud.bigquery").please_import("Client")()



from gcp_pal.utils import try_import

from gcp_pal.utils import log, ClientHandler, ModuleHandler, get_default_arg


class Storage:
    """
    Class for operating Google Cloud Storage
    """

    def __init__(self, path=None, bucket_name=None, project=None, location=None):
        self.project = project or get_default_arg("project")
        self.location = location or get_default_arg("location")
        self.fs_prefix = "gs://"

        self.path = path or self.fs_prefix
        if not self.path.startswith(self.fs_prefix):
            self.path = self.fs_prefix + self.path

        self.bucket_name = bucket_name
        if not self.bucket_name:
            self.bucket_name = self.path[len(self.fs_prefix) :].split("/")[0]
        else:
            if not path or path == self.fs_prefix:
                self.path = self.path + self.bucket_name

        self.base_path = self.path[len(self.fs_prefix) :]
        self.file_name = self.base_path[len(self.bucket_name) + 1 :]
        self.file_name = self.file_name if self.file_name != "" else None
        self.name = self.file_name or self.bucket_name

        self.bucket_path = self.fs_prefix + self.bucket_name
        if self.bucket_name == "":
            self.bucket_name = None
            self.bucket_path = None

        self.ref_type = self._ref_type()
        self.gcsfs = ModuleHandler("gcsfs").please_import(who_is_calling="Storage")
        self.GCSFileSystem = self.gcsfs.GCSFileSystem
        self.fs = ClientHandler(self.GCSFileSystem).get(
            project=self.project,
            default_location=self.location,
            cache_timeout=0,
        )

    def __repr__(self):
        """
        String representation of the class.

        Examples:
        - `Storage("bucket_name")` -> `Storage(gs://bucket_name)`
        """
        return f"Storage({self.path})"

    def _ref_type(self):
        """
        Get the reference type of the path.

        Returns:
        - str: Reference type of the path
        """
        self.is_file, self.is_bucket, self.is_project = False, False, False
        if not self.file_name and not self.bucket_name:
            self.is_project = True
            return "project"
        elif not self.file_name:
            self.is_bucket = True
            return "bucket"
        elif self.file_name:
            self.is_file = True
            return "file"
        raise ValueError("Invalid path")

    def ls(self, refresh=True):
        """
        List all files in the given path, or list all buckets in the project.

        Returns:
        - list: List of files in the bucket or list of buckets in the project.
        """
        if refresh:
            # It is not yet certain whether this will cause issues in large projects
            self.fs.invalidate_cache()
        return self.fs.ls(self.path)

    def download(self, local_path=None):
        """
        Download a file from the storage path to a local path.

        Args:
        - local_path (str): Local path where the file will be downloaded.
        """
        if not local_path:
            output = self.read()
        else:
            output = self.fs.get(self.path, local_path)
            log(f"Storage - Downloaded: {self.path} -> {local_path}")
        return output

    def glob(self, query=None):
        """
        List all files or directories matching the path regex.

        Args:
        - query (str): Path/query regex to match. Can also be provided in self.path.

        Returns:
        - list: List of matching files or directories.

        Examples:
        - `Storage("bucket_name").glob()` -> List all files in the bucket
        - `Storage("bucket_name/*/*").glob()` -> List all files in the bucket with two subdirectories
        """
        if query:
            return self.fs.glob(query)
        if not self.bucket_name:
            return self.ls()
        elif self.path.startswith(self.fs_prefix + "*"):
            raise ValueError("Wildcard is not supported for the bucket name yet.")
        return self.fs.glob(self.path)

    def exists(self, path=None):
        """
        Check if the file or directory exists in the path.

        Returns:
        - bool: Whether the file or directory exists.
        """
        self.fs.invalidate_cache()
        path = self._suffix_path(path)
        output = self.fs.exists(path)
        return output

    def isdir(self, path=None):
        """
        Check if the path is a directory.

        Returns:
        - bool: Whether the path is a directory.
        """
        self.fs.invalidate_cache()
        path = self._suffix_path(path)
        output = self.fs.isdir(path)
        try:
            object_info = self.fs.info(path)
        except FileNotFoundError:
            return output
        metadata = object_info.get("metadata", {})
        if object_info.get("type", "") == "directory":
            output = True
        elif metadata.get("arrow/gcsfs", "") == "directory":
            output = True
        elif object_info.get("name", "").endswith("/"):
            output = True
        return output

    def isfile(self, path=None):
        """
        Check if the path is a file.

        Returns:
        - bool: Whether the path is a file.
        """
        self.fs.invalidate_cache()
        path = self._suffix_path(path)
        output = self.fs.isfile(path)
        try:
            object_info = self.fs.info(path)
        except FileNotFoundError:
            return output
        metadata = object_info.get("metadata", {})
        if object_info.get("type", "") == "directory":
            output = False
        elif metadata.get("arrow/gcsfs", "") == "directory":
            output = False
        elif object_info.get("name", "").endswith("/"):
            output = False
        return output

    def mkdir(self, path=None, exist_ok=True):
        """
        Create a directory in the path.

        Args:
        - exist_ok (bool): Whether to ignore the error if the directory already exists.
        """
        path = self._suffix_path(path)
        output = self.fs.mkdir(path, exist_ok=exist_ok)
        log(f"Created directory: {path}")
        return output

    def mkdirs(self, path=None, exist_ok=True):
        """
        Create directories recursively in the path.

        Args:
        - exist_ok (bool): Whether to ignore the error if the directory already exists.
        """
        path = self._suffix_path(path)
        output = self.fs.mkdirs(path, exist_ok=exist_ok)
        log(f"Storage - Created directories: {path}")
        return output

    def create_bucket(self, bucket_name=None, exists_ok=True):
        """
        Create a bucket in the project.

        Args:
        - bucket_name (str): Name of the bucket to create.
        """
        if not bucket_name:
            bucket_name = self.bucket_name
        if not bucket_name:
            raise ValueError("Bucket name is required.")
        try:
            output = self.fs.mkdir(bucket_name)
        except self.gcsfs.retry.HttpError as e:
            exists = "Your previous request to create the named bucket succeeded and you already own it."
            if exists in str(e) and exists_ok:
                return
            raise e
        log(f"Storage - Created bucket: {bucket_name}")
        return output

    def create(self, path=None, exist_ok=True, bucket_name=None):
        """
        Alias for `create_bucket` or `mkdir`.

        Args:
        - path (str): Path to create
        - exist_ok (bool): Whether to ignore the error if the directory already exists.
        - bucket_name (str): Name of the bucket to create.
        """
        if bucket_name or self.is_bucket:
            return self.create_bucket(bucket_name, exists_ok=exist_ok)
        return self.mkdir(path, exist_ok=exist_ok)

    def walk(self):
        """
        Walk through the directory tree starting from the path.
        """
        return self.fs.walk(self.path)

    def rm(self, recursive=True):
        """
        Remove the file or directory in the path.

        Args:
        - recursive (bool): Whether to remove the directory recursively.
        """
        output = self.fs.rm(self.path, recursive=recursive)
        log(f"Storage - Removed: {self.path}")
        return output

    def rmdir(self, path=None):
        """
        Remove the directory in the path.

        Args:
        - path (str): Optional path from base path to remove.
        """
        path = self._suffix_path(path)
        path = self.bucket_name if self.is_bucket else path
        output = self.fs.rmdir(path)
        log(f"Storage - Removed: {path}")
        return output

    def remove(self, recursive=True):
        """
        Alias for `rm`.
        """
        return self.rm(recursive=recursive)

    def delete(self):
        """
        Deletes the file or directory in the path. If the path is a bucket, it deletes the bucket.

        Args:
        - recursive (bool): Whether to delete the directory recursively.
        """
        path = self.path
        if path.endswith("/"):
            path = path[:-1]
        try:
            output = self.rm()
        except FileNotFoundError:
            output = self.rmdir()
        return output

    def copy(self, destination_path, recursive=True):
        """
        Copy a file or directory from the source to the destination.

        Args:
        - destination_path (str): Destination path
        - recursive (bool): Whether to copy the directory recursively.
        """
        output = self.fs.copy(self.path, destination_path, recursive=recursive)
        log(f"Storage - Copied: {self.path} -> {destination_path}")
        return output

    def move(self, destination_path, recursive=True):
        """
        Move a file or directory from the source to the destination.

        Args:
        - destination_path (str): Destination path
        - recursive (bool): Whether to move the directory recursively.
        """
        output = self.fs.move(self.path, destination_path, recursive=recursive)
        log(f"Storage - Moved: {self.path} -> {destination_path}")
        return output

    def _upload(self, local_path=None, contents=None, recursive=True):
        """
        Upload a file or directory from the local path to the storage path.

        Args:
        - local_path (str): Local path from which to upload the file or directory.
        - contents (str): Contents to upload.
        - recursive (bool): Whether to upload the directory recursively.
        """
        if local_path:
            try:
                output = self.fs.put(local_path, self.path, recursive=recursive)
            except FileNotFoundError as e:
                self.create_bucket()
                output = self.fs.put(local_path, self.path, recursive=recursive)
            log(f"Storage - Uploaded: {local_path} -> {self.path}")
        elif contents:
            if isinstance(contents, str):
                mode = "w"
            else:
                mode = "wb"
            output = self.write(contents, mode=mode)
        return output

    def upload(self, local_path=None, contents=None, recursive=True):
        """
        Upload a file or directory from the local path to the storage path.

        Args:
        - local_path (str): Local path from which to upload the file or directory.
        - contents (str): Contents to upload.
        - recursive (bool): Whether to upload the directory recursively.
        """
        try:
            output = self._upload(local_path, contents, recursive)
        except FileNotFoundError as e:
            self.create_bucket()
            output = self._upload(local_path, contents, recursive)
        return output

    def put(self, local_path, recursive=True):
        """
        Alias for `upload`.
        """
        return self.upload(local_path, recursive=recursive)

    def open(self, mode="r"):
        """
        Open the storage path in the given mode.

        Args:
        - mode (str): Mode with which to open the file

        Returns:
        - file: File object
        """
        self.fs.invalidate_cache()
        return self.fs.open(self.path, mode=mode)

    def _suffix_path(self, path=None):
        """
        Append the path to the base path. Will not append if the path already starts with the fs_prefix.

        Args:
        - path (str): Path to append

        Returns:
        - str: Suffixed path

        Examples:
        >>> Storage()._suffix_path("file") -> "gs://file"
        >>> Storage("bucket_name")._suffix_path("file") -> "gs://bucket_name/file"
        >>> Storage("bucket_name/path")._suffix_path("gs://bucket_name") -> "gs://bucket_name/path"
        >>> Storage("bucket_name/path")._suffix_path() -> "gs://bucket_name"
        """
        if path is None:
            return self.path
        if path.startswith(self.fs_prefix):
            return path
        if path.startswith("/"):
            path = path[1:]
        if self.path.endswith("/"):
            return self.path + path
        return self.path + "/" + path

    def _write(self, data, path=None, mode="w", **kwargs):
        """
        Write data to the path.

        Args:
        - data: Data to write
        - path (str): Path to which to write the data
        - mode (str): Mode with which to write the file
        - kwargs: Additional arguments to pass to the write method
        """
        path = self._suffix_path(path)
        if path.endswith(".parquet") or path.endswith(".parquet/"):
            from gcp_pal import Parquet

            return Parquet(path).write(data, **kwargs)
        with self.fs.open(path, mode=mode) as f:
            f.write(data, **kwargs)
        log(f"Storage - Written: {path}")

    def write(self, data, path=None, mode="w", **kwargs):
        """
        Write data to the path.

        Args:
        - data: Data to write
        - path (str): Path to which to write the data
        - mode (str): Mode with which to write the file
        - kwargs: Additional arguments to pass to the write method
        """
        try:
            self._write(data, path, mode, **kwargs)
        except FileNotFoundError as e:
            self.create_bucket()
            self._write(data, path, mode, **kwargs)

    def read(self, path=None, **kwargs):
        """
        Read data from the path.

        Args:
        - path (str): Path to read
        - kwargs: Additional arguments to pass to the read

        Returns:
        - str: Data read from the path
        """
        self.fs.invalidate_cache()
        path = self._suffix_path(path)
        if path.endswith(".parquet") or path.endswith(".parquet/"):
            from gcp_pal import Parquet

            return Parquet(path).read(**kwargs)
        with self.fs.open(path, mode="r") as f:
            data = f.read()
        log(f"Storage - Read: {path}")
        return data

    def get(self, path=None, **kwargs):
        """
        Gets the object from the path as a reference.
        """
        try_import("google.cloud.storage", "Storage.get")
        from google.cloud.storage.blob import Blob
        from google.cloud.storage.bucket import Bucket

        path = self._suffix_path(path)
        bucket_name = path[len(self.fs_prefix) :].split("/")[0]
        path = path[len(bucket_name) + 1 :]
        bucket = Bucket(bucket_name, self.project)
        return Blob(path, bucket)


# if __name__ == "__main__":
#     import pyarrow as pa
#     import pyarrow.parquet as pq
#     import pandas as pd

#     success = {}
#     bucket_name = f"test_bucket_vita_1324"
#     Storage(bucket_name).create()

#     df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
#     partition_cols = ["a"]
#     file_name = f"gs://{bucket_name}/file.parquet"
#     table = pa.Table.from_pandas(df)
#     pq.write_to_dataset(
#         table, file_name, partition_cols=partition_cols, basename_template="{i}.parquet"
#     )
#     print(file_name)
#     print(Storage(file_name).isfile())
#     print(Storage(file_name).isfile("a=1"))
#     exit()

if __name__ == "__main__":
    print(Storage("bucket_name").bucket_name == "bucket_name")
    print(Storage("gs://bucket_name").bucket_name == "bucket_name")
    print(Storage("bucket_name/file").bucket_name == "bucket_name")
    print(Storage("gs://bucket_name/file").bucket_name == "bucket_name")
    print(Storage("gs://bucket_name").path == "gs://bucket_name")
    print(Storage("bucket_name").path == "gs://bucket_name")
    print(Storage("bucket_name/file").path == "gs://bucket_name/file")
    print(Storage("gs://bucket_name/file").path == "gs://bucket_name/file")
    print(Storage("gs://bucket_name/file").path == "gs://bucket_name/file")
    print(Storage(bucket_name="bucket").bucket_name == "bucket")
    print(Storage(bucket_name="bucket").path == "gs://bucket")
    print(Storage().path == "gs://")
    print(Storage().bucket_name is None)
    print(Storage().project)
    print(Storage().ls())
    print(Storage().glob())



import threading

from gcp_pal.utils import (
    log,
    get_all_kwargs,
    ClientHandler,
    ModuleHandler,
    get_default_arg,
)


class Dataplex:

    def __init__(
        self,
        path: str = None,
        lake: str = None,
        zone: str = None,
        asset: str = None,
        project: str = None,
        location: str = None,
    ):
        """
        Initializes the Dataplex object.

        Args:
        - path (str): The path of the object. The path is in the format `"lake/zone/asset"`, or `"lake/zone"`, `"lake"` or `None`.
                      The path will overwrite the lake, zone and asset attributes.
        - lake (str): The name of the lake.
        - zone (str): The name of the zone.
        - asset (str): The name of the asset.
        - project (str): The project ID. Default is the PROJECT environment variable or the default project from `gcloud config list`.
        - location (str): The location of the Dataplex resources. Default is `"europe-west2"`.
        """

        self.project = project or get_default_arg("project")
        self.location = location or get_default_arg("location")

        self.lake = lake
        self.zone = zone
        self.asset = asset

        if isinstance(path, str) and path.startswith("projects/"):
            path = self._convert_full_id_to_path(path)

        try:
            # e.g. path = "lake" or "lake/zone" or "lake/zone/asset"
            self.lake = path.split("/")[0]
        except:
            pass
        try:
            # e.g. path = "lake/zone" or "lake/zone/asset"
            self.zone = path.split("/")[1]
        except:
            pass
        try:
            # e.g. path = "lake/zone/asset"
            self.asset = path.split("/")[2]
        except:
            pass

        self.lake_id = self.lake
        self.zone_id = self.zone
        self.asset_id = self.asset

        self.type = self._get_type()
        self._verify_attributes()
        self.path = self._get_path()
        self.parent = self._get_parent()

        self.dataplex = ModuleHandler("google.cloud").please_import(
            "dataplex_v1", who_is_calling="Dataplex"
        )
        self.types = self.dataplex.types
        self.client = ClientHandler(self.dataplex.DataplexServiceClient).get()
        self.exceptions = ModuleHandler("google.api_core.exceptions").please_import(
            who_is_calling="Dataplex"
        )

    def _refresh_client(self):
        """
        Refresh the client. This is useful when the client caches some data and it needs to be refreshed.
        """
        self.client = ClientHandler(self.dataplex.DataplexServiceClient).get(
            force_refresh=True
        )

    def _get_type(self):
        """
        Get the type of the object. The type is either a lake, zone or an asset.

        Returns:
        - (str): The type of the object.
        """
        if not self.lake and not self.zone and not self.asset:
            return "project"
        elif self.lake and not self.zone and not self.asset:
            return "lake"
        elif self.lake and self.zone and not self.asset:
            return "zone"
        elif self.lake and self.zone and self.asset:
            return "asset"

    def _get_path(self):
        """
        Get the path of the object.

        Returns:
        - (str): The path of the object. Of the form "lake/zone/asset" or "lake/zone" or "lake" or None.
        """
        if self.type == "project":
            path = None
        elif self.type == "lake":
            path = self.lake
        elif self.type == "zone":
            path = f"{self.lake}/{self.zone}"
        elif self.type == "asset":
            path = f"{self.lake}/{self.zone}/{self.asset}"
        return path

    def _verify_attributes(self):
        """
        Verifies that the attributes are valid.
        For example, if the object is an asset, then it must have a lake, a zone and an asset.

        Raises:
        - ValueError: If the attributes are invalid.
        """
        if self.type == "asset" and not (self.lake and self.zone and self.asset):
            raise ValueError(
                "The object is an asset, so the following attributes must be provided: lake, zone, asset."
            )
        elif self.type == "zone" and not (self.lake and self.zone):
            raise ValueError(
                "The object is a zone, so the following attributes must be provided: lake, zone."
            )
        elif self.type == "lake" and not self.lake:
            raise ValueError(
                "The object is a lake, so the following attribute must be provided: lake."
            )

    def _get_parent(self):
        """
        Get the parent resource of the object.

        Returns:
        - (str): The parent resource of the object.
        """
        parent = f"projects/{self.project}/locations/{self.location}"
        if self.type == "lake":
            parent += ""
        elif self.type == "zone":
            parent += f"/lakes/{self.lake}"
        elif self.type == "asset":
            parent += f"/lakes/{self.lake}/zones/{self.zone}"
        return parent

    def _convert_full_id_to_path(self, full_id):
        """
        Converts the full resource ID to the path of the form "lake/zone/asset" or "lake/zone" or "lake" or None.

        Args:
        - full_id (str): The full resource ID.

        Returns:
        - (str): The path of the resource.
        """
        if not full_id:
            return None
        paths = full_id.split("/")[1::2][2:]
        path = "/".join(paths)
        return path

    def get_lake(self):
        """
        Retrieves the lake resource.

        Returns:
        - (dict): The lake resource.
        """
        name = f"{self.parent}/lakes/{self.lake}"
        got_lake = self.client.get_lake(name=name)
        return got_lake

    def get_zone(self):
        """
        Retrieves the zone resource.

        Returns:
        - (dict): The zone resource.
        """
        name = f"{self.parent}/zones/{self.zone}"
        got_zone = self.client.get_zone(name=name)
        return got_zone

    def get_asset(self):
        """
        Retrieves the asset resource.

        Returns:
        - (dict): The asset resource.
        """
        name = f"{self.parent}/assets/{self.asset}"
        got_asset = self.client.get_asset(name=name)
        return got_asset

    def get(self):
        """
        Retrieves the resource.

        Returns:
        - (dict): The resource.
        """
        got_method = {
            "lake": self.get_lake,
            "zone": self.get_zone,
            "asset": self.get_asset,
        }.get(self.type)
        got = got_method()
        return got

    def create_lake(
        self,
        display_name: str = None,
        description: str = None,
        labels: dict = None,
        metadata: dict = {},
        metastore_service: str = None,
        if_exists: str = "ignore",
        **kwargs,
    ):
        """
        Creates a lake resource.

        Args:
        - display_name (str): User-friendly display name of the lake. If not provided, it will be based on the lake_id.
        - description (str): The description of the lake resource.
        - labels (dict): The labels of the lake resource.
        - metadata (dict): The metadata of the lake resource.
        - metastore_service (str): The name of the metastore service.

        Returns:
        - (dict): The lake resource.
        """
        log(f"Dataplex - Creating lake '{self.lake_id}' [location: {self.location}]...")
        metastore = self.types.Lake.Metastore(service=metastore_service)
        lake = self.types.Lake(
            display_name=display_name,
            description=description,
            labels=labels,
            metastore=metastore,
        )
        try:
            created_lake = self.client.create_lake(
                parent=self.parent,
                lake_id=self.lake_id,
                lake=lake,
                timeout=600,
                metadata=metadata,
            )
        except self.exceptions.AlreadyExists as e:
            if if_exists.lower() == "ignore":
                log(f"Dataplex - Lake '{self.lake_id}' already exists.")
                return
            raise e
        result = created_lake.result()
        log(f"Dataplex - Lake '{self.lake_id}' created.")
        return result

    def create_zone(
        self,
        zone_type: str,  # "raw" or "curated"
        location_type: str = "single-region",  # "single-region"/"single" or "multi-region"/"multi"
        display_name: str = None,
        description: str = None,
        labels: dict = None,
        metadata: dict = {},
        if_exists: str = "ignore",
        **kwargs,
    ):
        """
        Creates a zone resource in the lake.

        Args:
        - zone_type (str): The type of the zone. Either `"raw"` or `"curated"`.
        - location_type (str): The location type of the zone. Either `"single-region"`/`"single"` or `"multi-region"`/`"multi"`.
        - display_name (str): User-friendly display name of the zone. If not provided, it will be based on the zone_id.
        - description (str): The description of the zone resource.
        - labels (dict): The labels of the zone resource.
        - metadata (dict): The metadata of the zone resource.
        - if_exists (str): If `"ignore"`, the operation will be ignored if the zone already exists.

        Returns:
        - (dict): The zone resource.
        """
        ResourceSpec = self.types.Zone.ResourceSpec
        LocationType = ResourceSpec.LocationType
        ZoneType = self.types.Zone.Type
        log(
            f"Dataplex - Creating zone '{self.zone_id}' [type: {zone_type}, location: {self.location} ({location_type})]..."
        )
        location_type = {
            "single-region": 1,
            "multi-region": 2,
            "single": 1,
            "multi": 2,
        }.get(location_type, 0)
        zone_type = {"raw": 1, "curated": 2}.get(zone_type, 0)
        location_type = LocationType(location_type)
        zone_type = ZoneType(zone_type)
        resource_spec = self.types.Zone.ResourceSpec(location_type=location_type)
        zone = self.types.Zone(
            display_name=display_name,
            description=description,
            labels=labels,
            type_=zone_type,
            resource_spec=resource_spec,
        )
        try:
            created_zone = self.client.create_zone(
                parent=self.parent,
                zone_id=self.zone_id,
                zone=zone,
                timeout=600,
                metadata=metadata,
            )
        except self.exceptions.AlreadyExists as e:
            if if_exists.lower() == "ignore":
                log(f"Dataplex - Zone '{self.zone_id}' already exists.")
                return
            raise e

        result = created_zone.result()
        log(f"Dataplex - Zone '{self.zone_id}' created.")
        return result

    def create_asset(
        self,
        asset_source: str,
        asset_type: str = None,
        display_name: str = None,
        description: str = None,
        labels: dict = None,
        metadata: dict = {},
        if_exists: str = "ignore",
        **kwargs,
    ):
        """
        Creates an asset resource in the zone in the lake.

        Args:
        - asset_source (str): The source of the asset. Either a bucket name or a dataset name.
        - asset_type (str): The type of the asset. Either `"storage"` or `"bigquery"`.
        - display_name (str): User-friendly display name of the asset. If not provided, it will be based on the asset_id.
        - description (str): The description of the asset resource.
        - labels (dict): The labels of the asset resource.
        - metadata (dict): The metadata of the asset resource.

        Returns:
        - (dict): The asset resource.
        """
        if not asset_type or asset_type not in ["storage", "bigquery"]:
            raise ValueError(
                "asset_type must be provided: Either 'storage' or 'bigquery'."
            )
        log(
            f"Dataplex - Creating asset '{self.asset_id}' [source: {asset_source}, type: {asset_type}]..."
        )
        ResourceSpec = self.types.Asset.ResourceSpec
        ResourceType = ResourceSpec.Type
        path_bit = {"storage": "buckets", "bigquery": "datasets"}.get(asset_type)
        asset_type = {"storage": 1, "bigquery": 2}.get(asset_type, 0)
        asset_type = ResourceType(asset_type)
        if not asset_source.startswith("projects/"):
            asset_source = f"projects/{self.project}/{path_bit}/{asset_source}"
        resource_spec = ResourceSpec(name=asset_source, type_=asset_type)
        asset = self.types.Asset(
            display_name=display_name,
            description=description,
            labels=labels,
            resource_spec=resource_spec,
        )
        try:
            created_asset = self.client.create_asset(
                parent=self.parent,
                asset_id=self.asset_id,
                asset=asset,
                timeout=600,
                metadata=metadata,
            )
        except self.exceptions.AlreadyExists as e:
            if if_exists.lower() == "ignore":
                log(f"Dataplex - Asset '{self.asset_id}' already exists.")
                return
            raise e
        result = created_asset.result()
        log(f"Dataplex - Asset '{self.asset_id}' created.")
        return result

    def create_parents(self, **kwargs):
        """
        Checks that the parent resources exist. If not, they will be created.

        Returns:
        - (bool): True if the operation was successful.
        """
        # We do not want to pass the redundant kwargs to the create methods of
        # the parent resources.
        redundant_kwargs = ["description", "display_name", "labels", "metadata"]
        kwargs = {k: v for k, v in kwargs.items() if k not in redundant_kwargs}
        if self.type == "lake":
            return True
        elif self.type == "zone":
            parent_lake = Dataplex(
                lake=self.lake,
                project=self.project,
                location=self.location,
            )
            if not parent_lake.exists():
                log(f"Dataplex - Parent lake '{self.lake}' does not exist.")
                parent_lake.create_lake(**kwargs)
        elif self.type == "asset":
            parent_lake = Dataplex(
                lake=self.lake,
                project=self.project,
                location=self.location,
            )
            if not parent_lake.exists():
                log(f"Dataplex - Parent lake '{self.lake}' does not exist.")
                parent_lake.create_lake(**kwargs)
            parent_zone = Dataplex(
                lake=self.lake,
                zone=self.zone,
                project=self.project,
                location=self.location,
            )
            if not parent_zone.exists():
                log(f"Dataplex - Parent zone '{self.zone}' does not exist.")
                parent_zone.create_zone(**kwargs)
        return True

    def create(
        self,
        asset_source: str,
        asset_type: str,
        zone_type: str = None,
        location_type: str = None,
        display_name: str = None,
        description: str = None,
        labels: dict = None,
        metadata: dict = {},
        **kwargs,
    ):
        """
        Creates the resource. If the lake doesn't exist, it will be created first. Then the zone and the assets will be created.

        Args:
        - asset_source (str): The source of the asset. Either a bucket name or a dataset name.
        - asset_type (str): The type of the asset. Either `"storage"` or `"bigquery"`. Has to be provided if the resource is a zone or an asset.
        - zone_type (str): The type of the zone. Either `"raw"` or `"curated"`. Has to be provided if the resource is a zone or an asset.
        - location_type (str): The location type of the zone. Either `"single-region"` or `"multi-region"`.
        - display_name (str): User-friendly display name of the resource. If not provided, it will be based on the resource_id.
        - description (str): The description of the resource.
        - labels (dict): The labels of the resource.
        - metadata (dict): The metadata of the resource.
        """
        all_kwargs = get_all_kwargs(locals())
        self.create_parents(**all_kwargs)
        if self.type == "lake":
            return self.create_lake(
                display_name=display_name,
                description=description,
                labels=labels,
                metadata=metadata,
            )
        elif self.type == "zone":
            return self.create_zone(
                zone_type=zone_type,
                location_type=location_type,
                display_name=display_name,
                description=description,
                labels=labels,
                metadata=metadata,
            )
        elif self.type == "asset":
            return self.create_asset(
                asset_source=asset_source,
                asset_type=asset_type,
                display_name=display_name,
                description=description,
                labels=labels,
                metadata=metadata,
            )

    def ls_lakes(self, name=None, full_id=False):
        """
        Lists the lakes in the project.

        Args:
        - name (str): Optional path of the lake.
        - full_id (bool): If True, returns the full resource ID of the lakes.

        Returns:
        - (list): The list of lakes.
        """
        parent = name or self.parent
        lakes = self.client.list_lakes(parent=parent)
        output = [lake.name for lake in lakes]
        if not full_id:
            output = [self._convert_full_id_to_path(name) for name in output]
        return output

    def ls_zones(self, name=None, full_id=False):
        """
        Lists the zones in the lake.

        Args:
        - name (str): Optional path of the zone.
        - full_id (bool): If True, returns the full resource ID of the zones. Otherwise returns the path of form "lake/zone".

        Returns:
        - (list): The list of zones.
        """
        output = []
        if self.type == "project":
            # Listing all zones in the project
            zones = []
            lakes = self.ls_lakes(full_id=True)
            for lake in lakes:
                lake = Dataplex(path=lake)
                zones = lake.ls_zones(full_id=True)
                output += zones
        else:
            parent = name or f"{self.parent}/lakes/{self.lake}"
            zones = self.client.list_zones(parent=parent)
            output = [zone.name for zone in zones]
        if not full_id:
            output = [self._convert_full_id_to_path(name) for name in output]
        return output

    def ls_assets(self, name=None, full_id=False):
        """
        Lists the assets in the zone.

        Args:
        - name (str): Optional path of the asset.
        - full_id (bool): If True, returns the full resource ID of the assets. Otherwise returns the path of form "lake/zone/asset".

        Returns:
        - (list): The list of assets.
        """
        output = []
        if self.type == "project":
            # Listing all assets in the project
            assets = []
            lakes = self.ls_lakes(full_id=True)
            for lake in lakes:
                lake = Dataplex(path=lake)
                zones = lake.ls_zones(full_id=True)
                for zone in zones:
                    zone = Dataplex(path=zone)
                    assets = zone.ls_assets(full_id=True)
                    output += assets
        elif self.type == "lake":
            # Listing all assets in the lake
            zones = self.ls_zones(full_id=True)
            for zone in zones:
                zone = Dataplex(path=zone)
                assets = zone.ls_assets(full_id=True)
                output += assets
        else:
            parent = name or f"{self.parent}/zones/{self.zone}"
            assets = self.client.list_assets(parent=parent)
            output = [asset.name for asset in assets]
        if not full_id:
            output = [self._convert_full_id_to_path(name) for name in output]
        return output

    def ls(self, level=None, full_id=False):
        """
        Lists the resources.

        Args:
        - level (str): The level of the resources to list. Either "lakes", "zones" or "assets".
        - full_id (bool): If True, returns the full resource ID of the resources.

        Returns:
        - (list): The list of resources.
        """
        if self.type == "asset":
            raise ValueError("The method 'Dataplex.ls' is not supported for assets.")
        if level:
            list_method = {
                "lakes": self.ls_lakes,
                "zones": self.ls_zones,
                "assets": self.ls_assets,
            }.get(level)
        else:
            list_method = {
                "project": self.ls_lakes,
                "lake": self.ls_zones,
                "zone": self.ls_assets,
            }.get(self.type)
        output = list_method(full_id=full_id)
        return output

    def exists(self):
        """
        Checks if the resource exists.

        Returns:
        - (bool): True if the resource exists, False otherwise.
        """
        try:
            self.get()
            return True
        except:
            return False

    def delete_lake(self, name: str = None, errors="ignore", wait_to_complete=True):
        """
        Deletes the lake resource.

        Args:
        - name (str): The full resource ID of the lake.

        Returns:
        - (dict): The response of the delete operation.
        """
        name = name or f"{self.parent}/lakes/{self.lake}"
        try:
            output = self.client.delete_lake(name=name)
        except self.exceptions.FailedPrecondition as e:
            msg = (
                f"Dataplex - Lake '{self.lake}' is not empty. Deleting all its zones..."
            )
            log(msg)
            lake = Dataplex(
                lake=self.lake, project=self.project, location=self.location
            )
            lake_assets = lake.ls_assets(full_id=True)
            for asset in lake_assets:
                Dataplex(path=asset).delete()
            zones = lake.ls_zones(full_id=True)
            for zone in zones:
                Dataplex(path=zone).delete()
            self._refresh_client()
            output = self.client.delete_lake(name=name)
        except self.exceptions.NotFound as e:
            if errors == "ignore":
                log(f"Dataplex - Lake '{self.lake}' does not exist to delete.")
                return
            raise e
        if wait_to_complete:
            output = output.result(timeout=600)
        log(f"Dataplex - Lake deleted: '{self.lake}'.")
        return output

    def delete_zone(self, name: str = None, errors="ignore", wait_to_complete=True):
        """
        Deletes the zone resource.

        Args:
        - name (str): The full resource ID of the zone.

        Returns:
        - (dict): The response of the delete operation.
        """
        name = name or f"{self.parent}/zones/{self.zone}"
        log(f"Dataplex - Deleting zone: '{self.path}'...")
        try:
            output = self.client.delete_zone(name=name)
        except self.exceptions.FailedPrecondition as e:
            msg = f"Dataplex - Zone '{self.path}' is not empty. Deleting all its assets..."
            log(msg)
            zone = Dataplex(
                lake=self.lake,
                zone=self.zone,
                project=self.project,
                location=self.location,
            )
            self._delete_parallel(zone.ls_assets(full_id=True))
            self._refresh_client()
            output = self.client.delete_zone(name=name)
        except self.exceptions.NotFound as e:
            if errors == "ignore":
                log(f"Dataplex - Zone '{self.path}' does not exist to delete.")
                return
            raise e
        if wait_to_complete:
            output = output.result(timeout=600)
        log(f"Dataplex - Zone deleted: '{self.path}'.")
        return output

    def delete_asset(self, name: str = None, errors="ignore", wait_to_complete=True):
        """
        Deletes the asset resource.

        Args:
        - name (str): The full resource ID of the asset.
        - wait_to_complete (bool): If True, waits until the operation is completed.

        Returns:
        - (dict): The response of the delete operation.
        """
        name = name or f"{self.parent}/assets/{self.asset}"
        log(f"Dataplex - Deleting asset: '{self.path}'...")
        try:
            output = self.client.delete_asset(name=name)
        except self.exceptions.NotFound as e:
            if errors == "ignore":
                log(f"Dataplex - Asset '{self.path}' does not exist to delete.")
                return
            raise e
        if wait_to_complete:
            output = output.result(timeout=600)
        log(f"Dataplex - Asset deleted: '{self.path}'.")
        return output

    def delete_lake(self, name: str = None, errors="ignore", wait_to_complete=True):
        """
        Deletes the lake resource using parallel deletion for assets and zones.

        Args:
        - name (str): The full resource ID of the lake.

        Returns:
        - (dict): The response of the delete operation.
        """
        name = name or f"{self.parent}/lakes/{self.lake}"
        try:
            output = self.client.delete_lake(name=name)
        except self.exceptions.FailedPrecondition:
            log(
                f"Dataplex - Lake '{self.lake}' is not empty. Deleting all its zones and assets..."
            )
            lake = Dataplex(
                lake=self.lake, project=self.project, location=self.location
            )

            # Deleting all assets in all zones in the lake
            self._delete_parallel(lake.ls_assets(full_id=True))

            # Deleting zones in the lake
            self._delete_parallel(lake.ls_zones(full_id=True))

            self._refresh_client()
            output = self.client.delete_lake(name=name)
        except self.exceptions.NotFound:
            if errors == "ignore":
                log(f"Dataplex - Lake '{self.lake}' does not exist to delete.")
                return
            raise
        if wait_to_complete:
            output = output.result(timeout=600)
        log(f"Dataplex - Lake deleted: '{self.lake}'.")
        return output

    def _delete_parallel(self, items):
        """
        Deletes items in parallel using threading.

        Args:
        - items (list): List of full resource IDs of assets or zones to delete.
        """
        threads = []
        for item in items:
            thread = threading.Thread(
                target=lambda item=item: Dataplex(path=item).delete()
            )
            thread.start()
            threads.append(thread)
        for thread in threads:
            thread.join()

    def delete(self, name: str = None, errors="ignore", wait_to_complete=True):
        """
        Deletes the resource.

        Args:
        - name (str): The full resource ID of the resource.
        - errors (str): If "ignore", the operation will be ignored if the resource does not exist.
        - wait_to_complete (bool): If True, waits until the operation is completed.

        Returns:
        - (dict): The response of the delete operation.
        """
        delete_method = {
            "lake": self.delete_lake,
            "zone": self.delete_zone,
            "asset": self.delete_asset,
        }.get(self.type)
        output = delete_method(
            name=name, errors=errors, wait_to_complete=wait_to_complete
        )
        return output

    def state(self):
        """
        Retrieves the state of the resource.

        Returns:
        - (str): The state of the resource.
        """
        state = self.get().state.name
        return state


if __name__ == "__main__":
    # Example: Create a Dataplex object
    from gcp_pal import BigQuery

    lake_name = "artem-lake1"
    zone_name = "artem-zone1"
    # Dataplex(path=lake_name).create_lake()
    # Dataplex(path=f"{lake_name}/{zone_name}").create_zone(
    #     zone_type="raw", location_type="single-region"
    # )
    # BigQuery(dataset="dataplex_dataset1").create()
    # BigQuery(dataset="dataplex_dataset2").create()
    # Dataplex(
    #     path=f"{lake_name}/{zone_name}/test-asset1",
    # ).create_asset(
    #     asset_source="dataplex_dataset1",
    #     asset_type="bigquery",
    # )
    # Dataplex(
    #     path=f"{lake_name}/{zone_name}/test-asset2",
    # ).create_asset(
    #     asset_source="dataplex_dataset2",
    #     asset_type="bigquery",
    # )

    Dataplex(lake_name, location="europe-west2").delete()



import os
import json

from gcp_pal.utils import log, ClientHandler, ModuleHandler, get_default_arg


class PubSub:
    """
    Class for pushing a payload to a topic

    Examples:
    - `PubSub("topic").publish("data")` -> Publish `"data"` to "topic"
    - `PubSub("topic").publish({"key": "value"})` -> Publish `{"key": "value"}` to "topic"
    """

    def __init__(self, topic: str, project=None):
        """
        Args:
        - topic (str): Name of the topic
        - project (str): Project ID
        """
        self.topic_id = topic
        self.project = project or get_default_arg("project")

        self.pubsub = ModuleHandler("google.cloud").please_import(
            "pubsub_v1", who_is_calling="PubSub"
        )
        self.publisher = ClientHandler(self.pubsub.PublisherClient).get()

        self.topic_path = self.publisher.topic_path(self.project, topic)

    def __repr__(self):
        return f"PubSub({self.topic_id})"

    def publish(self, data) -> None:
        if isinstance(data, dict):
            data = json.dumps(data)
        try:
            publish_future = self.publisher.publish(
                self.topic_path, data.encode("utf-8")
            )
            result = publish_future.result()
        except Exception as e:
            log(f"PubSub - An error occurred: {e}")


if __name__ == "__main__":
    PubSub("test_topic").publish("data")
    PubSub("test_topic").publish({"key": "value"})



from datetime import datetime, date, time
from gcp_pal.utils import (
    force_list,
    is_series,
    is_dataframe,
    log,
    reverse_dict,
    is_pyarrow_schema,
    is_bigquery_schema,
    is_python_schema,
    get_dict_items,
)


def get_equivalent_schema_dict(target):
    """
    Get the equivalent schema dictionary in the target system.

    Args:
    - target (str): The target system.

    Returns:
    - dict: The equivalent schema dictionary in the target system.
    """
    if target == "str":
        return {
            "int": "int",
            "float": "float",
            "str": "str",
            "bool": "bool",
            "timestamp": "timestamp",
            "date": "date",
            "time": "time",
            "datetime": "datetime",
            "bytes": "bytes",
            "array": "array",
            "struct": "struct",
            "null": "null",
        }
    elif target == "python":
        return {
            "int": int,
            "float": float,
            "str": str,
            "bool": bool,
            "timestamp": datetime,
            "date": date,
            "time": time,
            "datetime": datetime,
            "bytes": bytes,
            "array": list,
            "struct": dict,
            "null": type(None),
        }
    elif target == "bigquery":
        return {
            "null": "BOOLEAN",
            "int": "INTEGER",
            "float": "FLOAT",
            "str": "STRING",
            "bool": "BOOLEAN",
            "timestamp": "TIMESTAMP",
            "date": "DATE",
            "time": "TIME",
            "datetime": "DATETIME",
            "bytes": "BYTES",
            "array": "ARRAY",
            "struct": "STRUCT",
        }
    elif target == "pandas":
        return {
            "int": "Int64",
            "float": "Float64",
            "str": "string",
            "bool": "boolean",
            "timestamp": "datetime64[ns]",
            "date": "datetime64[ns]",
            "time": "datetime64[ns]",
            "datetime": "datetime64[ns]",
            "bytes": "bytes",
            "array": "list",
            "struct": "struct",
            "null": "object",
        }
    elif target == "pyarrow":
        return {
            "int": "int64",
            "float": "double",
            "str": "string",
            "bool": "bool",
            "timestamp": "timestamp[ns]",
            "date": "date32",
            "time": "time64[ns]",
            "datetime": "timestamp[ns]",
            "bytes": "binary",
            "array": "list",
            "struct": "struct",
            "null": "null",
        }
    else:
        raise ValueError(f"Unsupported target system: {target}")


ALL_SUPPORTED_SCHEMA_TYPES = ["bigquery", "str", "python", "pandas", "pyarrow"]


def compute_type_matches(schema: list, target_schema: list):
    """
    Compute the type matches between two schema lists.

    Args:
    - schema (list): The input schema values.
    - target_schema (list): The target schema values against which to compare.

    Returns:
    - int: The number of type matches.
    """
    type_matches = 0
    for value in schema:
        if value in target_schema:
            type_matches += 1
    return type_matches


def get_matching_schema_type(schema):
    """
    Get the matching schema type for the schema.

    Args:
    - schema (dict): The schema.

    Returns:
    - str: The matching schema type.
    """
    if not isinstance(schema, list):
        schema_values = get_dict_items(schema, "value")
    else:
        schema_values = schema
    type_matches = {}
    for target in ALL_SUPPORTED_SCHEMA_TYPES:
        target_schema = get_equivalent_schema_dict(target)
        target_values = get_dict_items(target_schema, "value")
        type_matches[target] = compute_type_matches(schema_values, target_values)

    matching_values = type_matches.values()
    matching_type = max(type_matches, key=type_matches.get)
    if sum(matching_values) == 0:
        log("Schema - No matching schema type found.")
        return None
    elif sum([x == len(schema) for x in matching_values]) > 1:
        log("Schema - Multiple matching schema types found.")
        return None
    matching_type = max(type_matches, key=type_matches.get)
    if type_matches[matching_type] != len(schema_values):
        log("Schema - Not all schema types matched.")
        return None
    return matching_type


def get_equivalent_schema_type(schema_type, target="bigquery", direction="forward"):
    """
    Get the equivalent schema type in the target system.

    Args:
    - schema_type (str): The schema type.
    - target (str): The target system.
    - direction (str): The direction of the conversion. This is used for converting between systems.

    Returns:
    - str: The equivalent schema type in the target system.
    """
    schema_dict = get_equivalent_schema_dict(target, direction)
    if direction == "reverse":
        schema_dict = reverse_dict(schema_dict)
    if schema_type in schema_dict:
        return schema_dict[schema_type]
    else:
        raise ValueError(f"Unsupported schema type: {schema_type}")


def get_equivalent_schema(
    schema, origin=None, target=None, schema_from=None, schema_to=None
):
    """
    Get the equivalent schema in the target system.

    Args:
    - schema (dict): The schema to convert.
    - origin (str): The origin system.
    - target (str): The target system.

    Returns:
    - dict: The equivalent schema in the target system.
    """
    if origin is None:
        origin = "str"
    # In a parallel universe where Python is a functional language...
    # origin -> target = (str -> origin) -> target = (origin <- str) -> target
    equivalent_schema = {}
    if schema_from is None:
        schema_from = reverse_dict(get_equivalent_schema_dict(origin))
    if schema_to is None:
        schema_to = get_equivalent_schema_dict(target)
    for col, col_type in schema.items():
        if isinstance(col_type, dict):
            type_to = get_equivalent_schema(
                col_type, origin, target, schema_from, schema_to
            )
        else:
            type_from = schema_from[col_type]
            type_to = schema_to[type_from]
        equivalent_schema[col] = type_to
    return equivalent_schema


def dtype_str_to_type(dtype_str, target="str"):
    """
    Map a string representation of a Pandas dtype to its corresponding Python or str type
    or return the string if it's a Pandas-specific type.

    Args:
    - dtype_str (str): The string representation of the dtype.

    Returns:
    - type: The Python type corresponding to the dtype string.
    """
    if target == "python":
        target_types = {
            "int": int,
            "int64": int,
            "float": float,
            "float64": float,
            "str": str,
            "bool": bool,
            "object": str,
            "datetime64[ns]": datetime,
        }
    elif target == "str":
        target_types = {
            "int": "int",
            "int64": "int",
            "Int64": "int",
            "float": "float",
            "float64": "float",
            "Float64": "float",
            "str": "str",
            "string": "str",
            "bool": "bool",
            "boolean": "bool",
            "object": "str",
            "datetime64[ns]": "datetime",
        }
    return target_types.get(dtype_str, dtype_str)


def dict_to_bigquery_fields(schema_dict):
    """
    Convert a dictionary to a BigQuery schema.

    Args:
    - schema_dict (dict): The dictionary to convert.

    Returns:
    - list[bigquery.SchemaField]: The BigQuery schema.
    """
    from google.cloud import bigquery

    schema = []
    for col, col_type in schema_dict.items():
        if isinstance(col_type, dict):
            schema.append(
                bigquery.SchemaField(
                    col, "RECORD", fields=dict_to_bigquery_fields(col_type)
                )
            )
        else:
            schema.append(bigquery.SchemaField(col, col_type))

    return schema


def bigquery_fields_to_dict(schema):
    """
    Convert a schema to a list of columns.

    Args:
    - schema: list, the schema to convert.

    Returns:
    - (dict): The columns schema of the form
              {column_name: column_type, column_name: {nested_column_name: nested_column_type}}
    """
    columns = {}
    for field in schema:
        name = field.name
        fields = field.fields
        if fields:
            fields = bigquery_fields_to_dict(fields)
        columns[name] = fields
        if not fields:
            columns[name] = field.field_type
    return columns


def bigquery_fields_dict_to_dict(schema_fields_dict):
    """
    Convert BigQuery schema fields dictionary to a string dictionary.

    Args:
    - schema_fields_dict (dict[bigquery.SchemaField]): The BigQuery schema fields dictionary.

    Returns:
    - dict: The dictionary representation of the schema.

    Examples:
    >>> bigquery_fields_dict_to_dict({"a": bigquery.SchemaField("a", "INTEGER"),
                                      "b": bigquery.SchemaField("b", "STRING")})
        {"a": "INTEGER", "b": "STRING"}
    """
    schema_dict = {}
    for col, col_type in schema_fields_dict.items():
        if isinstance(col_type, dict):
            schema_dict[col] = bigquery_fields_dict_to_dict(col_type)
        elif isinstance(col_type, str):
            schema_dict[col] = col_type
        else:
            schema_dict[col] = col_type.field_type
    return schema_dict


def bigquery_to_dict(schema):
    """
    Convert a BigQuery schema to a dictionary.

    Args:
    - schema (list[bigquery.SchemaField]): The BigQuery schema.

    Returns:
    - dict[str, str]: The dictionary representation of the schema with string types.
    """
    if isinstance(schema, dict):
        output = bigquery_fields_dict_to_dict(schema)
    else:
        output = bigquery_fields_to_dict(schema)
    return output


def dict_to_pyarrow_fields(schema_dict):
    """
    Convert a dictionary to a PyArrow schema.

    Args:
    - schema_dict (dict): The dictionary to convert.

    Returns:
    - pyarrow.Schema: The PyArrow schema.
    """
    import pyarrow as pa

    fields = []
    if not isinstance(schema_dict, dict):
        return getattr(pa, schema_dict)()
    for col, col_type in schema_dict.items():
        if isinstance(col_type, dict):
            fields.append(pa.field(col, pa.struct(dict_to_pyarrow_fields(col_type))))
        elif isinstance(col_type, list):
            fields.append(pa.field(col, pa.list_(dict_to_pyarrow_fields(col_type[0]))))
        else:
            fields.append(pa.field(col, col_type))
    schema = pa.schema(fields)
    return schema


def pyarrow_fields_dict_to_dict(schema_fields_dict):
    """
    Convert PyArrow schema fields dictionary to a string dictionary.

    Args:
    - schema_fields_dict (dict[pyarrow.SchemaField]): The PyArrow schema fields dictionary.

    Returns:
    - dict: The dictionary representation of the schema.

    Examples:
    >>> pyarrow_fields_dict_to_dict({"a": pa.int64(), "b": pa.string(), "c": {"d": pa.float64()}})
    {"a": "int64", "b": "string", "c": {"d": "double"}}
    """
    schema_dict = {}
    for col, col_type in schema_fields_dict.items():
        if isinstance(col_type, dict):
            schema_dict[col] = pyarrow_fields_dict_to_dict(col_type)
        else:
            schema_dict[col] = str(col_type)
    return schema_dict


def pyarrow_fields_to_dict(schema_fields):
    """
    Convert PyArrow schema fields to a dictionary.

    Args:
    - schema_fields (pyarrow.schema): The PyArrow schema.

    Returns:
    - dict: The dictionary representation of the schema.
    """
    schema_dict = {}
    for field in schema_fields:
        field_type = field.type
        # Check if nested
        if field_type.num_fields > 0:
            schema_dict[field.name] = pyarrow_fields_to_dict(field_type)
        else:
            schema_dict[field.name] = str(field_type)
    return schema_dict


def pyarrow_to_dict(schema):
    """
    Convert a PyArrow schema to a dictionary.

    Args:
    - schema (pyarrow.Schema): The PyArrow schema as a dictionary or pa.schema object.

    Returns:
    - dict[str, str]: The dictionary representation of the schema with string types.
    """
    if isinstance(schema, dict):
        output = pyarrow_fields_dict_to_dict(schema)
    else:
        output = pyarrow_fields_to_dict(schema)
    return output


def type_to_str(schema_dict):
    """
    Convert a dictionary to a string representation of the schema.

    Args:
    - schema_dict (dict): The dictionary to convert.

    Returns:
    - str: The string representation of the schema.
    """
    schema_str = {}
    if schema_dict == []:
        return []
    if not isinstance(schema_dict, dict):
        return schema_dict.__name__
    for col, col_type in schema_dict.items():
        if isinstance(col_type, dict):
            schema_str[col] = {str(k): type_to_str(v) for k, v in col_type.items()}
        elif isinstance(col_type, list):
            schema_str[col] = [type_to_str(v) for v in col_type]
        else:
            schema_str[col] = col_type.__name__
    return schema_str


def ensure_types(df, types_dict):
    """
    Ensure the types of the columns of the df.
    """
    for c, c_type in types_dict.items():
        if c in df.columns:
            df[c] = df[c].astype(c_type)
    return df


def enforce_schema_on_series(series, schema):
    """
    Enforce a schema on a pandas Series.
    """
    if isinstance(schema, (type, str)):
        return series.astype(schema)
    elif isinstance(schema, type(lambda x: x)):
        return series.apply(schema)
    elif callable(schema):
        return schema(series)
    elif isinstance(schema, dict):
        return series.map(lambda x: schema.get(x, x))
    else:
        raise TypeError("Unsupported schema type.")


def enforce_schema_on_list(lst, schema):
    """
    Enforce a schema on a list.
    """
    if callable(schema):
        return [schema(x) for x in lst]
    elif isinstance(schema, dict):
        return [schema.get(x, x) for x in lst]
    elif isinstance(schema, type):
        return [schema(x) for x in lst]
    elif isinstance(schema, str):
        schema = dtype_str_to_type(schema, target="python")
        return [schema(x) for x in lst]
    else:
        raise TypeError("Unsupported schema type.")


def enforce_one_schema(data, col_schema):
    """
    Enforce a schema on a dataframe column or a list of data.

    Args:
    - data (Series|List): The data to enforce the schema on.
    - col_schema (type|dict|callable|list): The schema to enforce.

    Returns:
    - Series|List: The data with the schema enforced.
    """

    # Helper function to ensure data is a list
    def force_list(data):
        return data if isinstance(data, list) else [data]

    if isinstance(col_schema, list):
        # Attempt to enforce each schema in the list until one succeeds
        for schema in col_schema:
            try:
                return enforce_one_schema(data, schema)
            except Exception:
                continue
        else:
            raise ValueError(f"Could not enforce schema {col_schema} on {data}")
    else:
        if is_series(data):
            return enforce_schema_on_series(data, col_schema)
        else:
            data = force_list(data)
            return enforce_schema_on_list(data, col_schema)


def enforce_schema(df, schema={}, dtypes={}, errors="raise"):
    """
    Enforce a schema on a dataframe or dictionary.

    Args:
    - df (DataFrame|dict): The data to enforce the schema on.
    - schema (dict): The schema to enforce.
    - dtypes (dict): The object types to enforce. The schema will override these.
    - errors (str): How to handle errors.

    Returns:
    - DataFrame|dict: The data with the schema enforced.
    """
    schema = {**dtypes, **schema}  # schema takes precedence over dtypes
    if schema == {}:
        return df
    for col, col_schema in schema.items():
        if col not in df:
            df[col] = None
        try:
            df[col] = enforce_one_schema(df[col], col_schema)
        except Exception as e:
            log(f"Schema - Error enforcing schema {col_schema} on {col}: {e}")
            if errors == "raise":
                raise e
    return df


def infer_schema(data, schema_type="python"):
    """
    Infer a schema from a dataframe or dictionary.

    Args:
    - data (DataFrame|dict): The data to infer the schema from.

    Returns:
    - dict: The inferred schema.
    """
    schema = {}
    if is_dataframe(data):
        data = data.convert_dtypes(convert_integer=False)
        for col, dtype in data.dtypes.items():
            schema[col] = dtype_str_to_type(str(dtype))
    elif isinstance(data, dict):
        for col, values in data.items():
            if isinstance(values, dict):
                schema[col] = infer_schema(values)
            elif values or isinstance(values, bool) or values is None:
                non_null_values = [v for v in force_list(values) if v is not None]
                value = non_null_values[0] if non_null_values else None
                if value is None:
                    schema[col] = "null"
                elif isinstance(value, bool):
                    schema[col] = "bool"
                elif isinstance(value, int):
                    schema[col] = "int"
                elif isinstance(value, float):
                    schema[col] = "float"
                elif isinstance(value, str):
                    schema[col] = "str"
                else:
                    schema[col] = type(value).__name__
    return schema


class Schema:
    """
    A class for for bridging the gap between different schema representations of similar data.

    Args:
    - schema (dict): The input schema dictionary.
    - schema_type (str): The schema type. Options: "str", "python", "pandas", "pyarrow".

    Returns:
    - Schema: The schema object.

    Examples:
    - `Schema({"a": "int", "b": "str"}, schema_type="str").bigquery()` -> BigQuery schema
    - `Schema({"a": "int", "b": "str"}, schema_type="str").python()` -> Python schema
    - `Schema({"a": "int", "b": "str"}, schema_type="str").pandas()` -> Pandas schema
    - `Schema({"a": "int", "b": "str"}, schema_type="str").pyarrow()` -> PyArrow schema
    - `Schema({"a": "int", "b": "str"}, schema_type="str").str()` -> String schema
    """

    def __init__(self, input: dict = {}, schema_type: str = None, is_data=False):
        self.is_data = is_data if not is_dataframe(input) else True
        if self.is_data and isinstance(input, list) and isinstance(input[0], dict):
            input = input[0]
        self.input_schema = input
        self.schema = input
        self.schema_type = schema_type or self.infer_schema_type()

        # Now the goal is to convert whatever schema into a dictionary of Python types
        if is_dataframe(input) or self.is_data:
            self.infer_schema()
        if self.schema_type == "bigquery":
            self.schema = bigquery_to_dict(input)
        elif self.schema_type == "pyarrow":
            self.schema = pyarrow_to_dict(input)
        if self.schema_type != "str":
            self.convert_schema_to_str()
        # Now the schema is a dictionary of Python types

    def __repr__(self):
        return f"Schema({self.schema})"

    def convert_schema_to_str(self):
        """
        Convert the schema to a Python dictionary.
        """
        self.schema = get_equivalent_schema(self.schema, self.schema_type, "str")
        self.schema_type = "str"

    def infer_schema_type(self):
        """
        Infer the schema type from the input.

        Returns:
        - str: The schema type.

        Examples:
        - `Schema({"a": int, "b": str}).infer_schema_type()` -> "python"
        - `Schema({"a": "int", "b": "str"}).infer_schema_type()` -> "str"
        - `Schema({"a": "int64", "b": "str"}).infer_schema_type()` -> "pandas"
        - `Schema(pa.schema([pa.field("a", pa.int64()), pa.field("b", pa.string())])).infer_schema_type()` -> "pyarrow"
        - `Schema([bigquery.SchemaField("a", "INTEGER"), bigquery.SchemaField("b", "STRING")]).infer_schema_type()` -> "bigquery"
        """
        if not self.is_data and is_pyarrow_schema(self.input_schema):
            return "pyarrow"
        if not self.is_data and is_bigquery_schema(self.input_schema):
            return "bigquery"
        if not self.is_data and is_python_schema(self.input_schema):
            return "python"
        if self.is_data or is_dataframe(self.input_schema):
            # We return None here because self.input_schema is not a schema dictionary
            return None
        if not isinstance(self.input_schema, dict):
            return None
        matching_type = get_matching_schema_type(self.input_schema)

        return matching_type

    def infer_schema(self) -> "Schema":
        """
        Infer the schema from the input.

        Returns:
        - Schema: The schema object with the schema set to the inferred schema.
        """
        if self.schema_type == "str":
            return self
        self.schema = infer_schema(self.input_schema, self.schema_type)
        self.schema_type = "str"
        return self

    def bigquery(self) -> dict:
        if self.schema_type == "bigquery":
            return self.schema
        output_schema = get_equivalent_schema(self.schema, self.schema_type, "bigquery")
        output_schema = dict_to_bigquery_fields(output_schema)
        return output_schema

    def python(self) -> dict:
        if self.schema_type == "python":
            return self.schema
        output_schema = get_equivalent_schema(self.schema, self.schema_type, "python")
        return output_schema

    def pandas(self) -> dict:
        if self.schema_type == "pandas":
            return self.schema
        output_schema = get_equivalent_schema(self.schema, self.schema_type, "pandas")
        return output_schema

    def pyarrow(self) -> dict:
        if self.schema_type == "pyarrow":
            return self.schema
        output_schema = get_equivalent_schema(self.schema, self.schema_type, "pyarrow")
        output_schema = dict_to_pyarrow_fields(output_schema)
        return output_schema

    def str(self) -> str:
        if self.schema_type == "str":
            return self.schema
        output_schema = type_to_str(self.schema)
        return output_schema


if __name__ == "__main__":
    import pandas as pd
    import pyarrow as pa

    success = {}
    df = pd.DataFrame(
        {
            "a": [1, 2, 3],
            "b": ["a", "b", "c"],
            "c": [1.0, 2.0, 3.0],
            "date": [datetime.now() for _ in range(3)],
        }
    )

    # string
    schema = Schema(df).str()
    str_schema = {"a": "int", "b": "str", "c": "float", "date": "datetime"}
    success[0] = schema == str_schema

    print(schema)

    exit()

if __name__ == "__main__":
    import pandas as pd
    import pyarrow as pa

    pyarrow_schema = {"a": pa.int64(), "b": pa.string(), "c": {"d": pa.float64()}}

    # 1. pyarrow -> str
    schema = Schema(pyarrow_schema).str()
    str_schema = {"a": "int", "b": "str", "c": "float"}
    print(str_schema)
    exit()

    inferred_schema = Schema(df, schema_type="pandas").infer_schema()
    python_schema = inferred_schema.python()
    bigquery_schema = inferred_schema.bigquery()
    str_schema = inferred_schema.str()
    pyarrow_schema = inferred_schema.pyarrow()
    print("Inferred schema:", inferred_schema.schema)
    print("Python schema:", python_schema)
    print("BigQuery schema:", bigquery_schema)
    print("String schema:", str_schema)
    print("PyArrow schema:", pyarrow_schema)
    exit()



from gcp_pal.utils import log, ModuleHandler, get_default_arg


class Docker:

    _client = None

    def __init__(
        self,
        name,
        project=None,
        tag="latest",
        destination=None,
        repository="docker",
        location=None,
    ):
        """
        Initialize a Docker object.

        Args:
        - name (str): The name of the Docker image.
        - project (str): The GCP project ID. Defaults to the PROJECT environment variable or the default auth project.
        - tag (str): The tag of the Docker image. Defaults to 'latest'.
        - destination (str): The destination of the pushed image. Defaults to gcr.io/{project}/{name}:{tag}.
        - repository (str): The name of the Artifact Registry repository. Defaults to 'docker'.
        """
        self.docker = ModuleHandler("docker").please_import(who_is_calling="Docker")
        if self._client is None:
            self.client = self.docker.from_env()
            Docker._client = self.client
        else:
            self.client = self._client

        self.project = project or get_default_arg("project")
        self.location = location or get_default_arg("location")
        self.name = name
        self.tag = tag
        self.repository = repository
        gcr_dest = f"gcr.io/{self.project}/{self.name}:{self.tag}"
        ar_dest = f"{self.location}-docker.pkg.dev/{self.project}/{self.repository}/{self.name}:{self.tag}"
        self.default_dest = gcr_dest
        self.destination = destination or self.default_dest

    def build(self, path=".", dockerfile="Dockerfile", verbose=False, **kwargs):
        """
        Build a Docker image from a Dockerfile.

        Args:
        - path (str): The path to the build context.
        - dockerfile (str): The relative path to the Dockerfile from the context path.
        - verbose (bool): Whether to stream the output of the build command.
        - kwargs: Additional arguments to pass to the Docker images build command.

        Returns:
        - None
        """
        log(f"Docker - Building image '{self.name}:{self.tag}' from {dockerfile}...")
        _, output = self.client.images.build(
            path=path,
            tag=self.destination,
            dockerfile=dockerfile,
            **kwargs,
        )
        if verbose:
            for line in output:
                log(line)
        # Check for success
        for line in output:
            stream = line.get("stream")
            if stream is not None and "Successfully built" in stream:
                log(f"Docker - Image '{self.name}:{self.tag}' built successfully.")
                return
        log(f"Docker - Image '{self.name}:{self.tag}' failed to build.")
        return

    def push(self, verbose=False, destination=None, **kwargs):
        """
        Push the built image to GCR.

        Args:
        - verbose (bool): Whether to stream the output of the push command.
        - destination (str): The destination of the pushed image. Defaults to gcr.io/{project}/{name}:{tag}.
        - kwargs: Additional arguments to pass to the Docker images push command.

        Returns:
        - (str) The destination of the pushed image (or None if an error occurred)
        """
        destination = destination or self.destination
        log(f"Docker - Pushing image to {destination}...")
        stream = True if verbose else False
        output = self.client.images.push(
            destination,
            stream=stream,
            decode=True,
            **kwargs,
        )
        # Check for success
        for line in output:
            if "error" in line:
                log(f"Docker - Error: {line}")
                return
        log(f"Docker - Pushed '{self.name}' -> {destination}.")
        return destination

    def build_and_push(
        self, path=".", dockerfile="Dockerfile", verbose=False, **kwargs
    ):
        """
        Build and push a Docker image to GCR.

        Args:
        - path (str): The path to the build context.
        - dockerfile (str): The relative path to the Dockerfile from the context path.
        - verbose (bool): Whether to stream the output of the build and push commands.
        - kwargs: Additional arguments to pass to the Docker images build and push commands.

        Returns:
        - (str) The destination of the pushed image (or None if an error occurred)
        """
        self.build(path=path, dockerfile=dockerfile, verbose=verbose, **kwargs)
        output = self.push(verbose=verbose, **kwargs)
        return output


# Example usage
if __name__ == "__main__":
    context = "samples/cloud_run"
    Docker("test-app").build(context)
    Docker("test-app").push()



from __future__ import annotations

import os
from gcp_pal.utils import try_import

from gcp_pal.schema import (
    Schema,
    dict_to_bigquery_fields,
    dict_to_bigquery_fields,
    bigquery_fields_to_dict,
)
from gcp_pal.utils import (
    is_dataframe,
    orient_dict,
    log,
    ClientHandler,
    ModuleHandler,
    get_default_arg,
)


class SQLBuilder:
    """
    Class for safely building SQL queries for use with Google BigQuery.
    """

    ALLOWED_OPERATIONS = {"=", ">", "<", ">=", "<=", "IN", "LIKE"}

    def __init__(self, table_name: str):
        self.table_name = table_name
        self.query_parts = ["FROM", f"`{table_name}`"]
        self.parameters = {}  # Holds parameters to prevent SQL injection

    def build(self) -> str:
        """
        Return the final query string and parameters.
        """
        sql_query = " ".join(self.query_parts), self.parameters

        # Reset query parts and parameters
        self.query_parts = ["FROM", f"`{self.table_name}`"]
        self.parameters = {}
        return sql_query

    def select(self, columns=None) -> "SQLBuilder":
        """
        Build a SELECT query.
        """
        if columns is None:
            columns = "*"
        if isinstance(columns, str):
            columns = [columns]
        columns = [f"`{x}`" for x in columns if x != "*"] or ["*"]
        select_clause = "SELECT " + ", ".join(columns)
        self.query_parts.insert(0, select_clause)  # Insert at the beginning
        return self

    def where(self, filters: list[tuple]) -> "SQLBuilder":
        """
        Build a WHERE clause with safe parameter handling.
        """
        if not filters:
            return self

        where_clause, params = self._where(filters)
        if where_clause:
            self.query_parts.append("WHERE " + where_clause)
            self.parameters.update(params)
        return self

    def limit(self, limit: int) -> "SQLBuilder":
        """
        Build a LIMIT clause.
        """
        if limit is not None and isinstance(limit, int) and limit > 0:
            self.query_parts.append(f"LIMIT {limit}")
        return self

    def _where(self, filters: list[tuple]) -> tuple:
        """
        Helper to build a WHERE clause safely.
        """
        conditions = []
        params = {}
        for i, filter in enumerate(filters):
            if isinstance(filter, dict):
                col, value = list(filter.items())[0]
                op = "IN"
            else:
                col, op, value = filter
            self._check_illegal_filters(col, op, value)
            param_name = f"param_{i}"
            if isinstance(value, list):
                placeholders = ", ".join(
                    [f"@{param_name}_{j}" for j, _ in enumerate(value)]
                )
                conditions.append(f"`{col}` {op} ({placeholders})")
                for j, v in enumerate(value):
                    params[f"{param_name}_{j}"] = v
            else:
                conditions.append(f"`{col}` {op} @{param_name}")
                params[param_name] = value
        return " AND ".join(conditions), params

    def _check_illegal_filters(self, col, op, value):
        """
        Helper to check for illegal characters in filters.
        """
        op = op.upper()
        if op not in self.ALLOWED_OPERATIONS:
            raise ValueError(f"Filter operator not allowed: {op}")
        if "`" in col:
            raise ValueError("Column or value contains illegal character: `")
        if "--" in col:
            raise ValueError("Column or value contains illegal character: --")
        if isinstance(value, str) and "--" in value:
            raise ValueError("Column or value contains illegal character: --")
        if isinstance(value, str) and "`" in value:
            raise ValueError("Column or value contains illegal character: `")


class BigQuery:
    """
    Class for operating Google BigQuery.
    """

    def __init__(self, table=None, dataset=None, project=None, location=None):
        """
        Initializes the BigQuery client.

        Args:
        - table (str): Table or table ID for the BigQuery service.
        - dataset (str): Dataset or dataset ID for the BigQuery service.
        - project (str): Project ID for the BigQuery service.
        - location (str): Location for the BigQuery service (default: "europe-west2").

        Examples:
        - `BigQuery().query("SELECT * FROM project.dataset.table")` -> Executes a query and returns the results.
        - `BigQuery("dataset.table").insert(data)` -> Inserts data into the specified BigQuery table.
        - `BigQuery("dataset.new_table").create_table(schema=schema)` -> Creates a new table with the specified schema.

        Storage-equivalent examples:
        - `BigQuery().ls()` -> Lists all datasets in the project.
        - `BigQuery("dataset").ls()` -> Lists all tables in the specified dataset.
        - `BigQuery("dataset.table").read()` -> Reads the entire table.
        - `BigQuery("dataset.table").write(data)` -> Writes data to the specified table.
        - `BigQuery("dataset.table").delete()` -> Deletes the specified table.
        - `BigQuery("dataset").delete()` -> Deletes the specified dataset.
        """
        self.table = table
        self.dataset = dataset
        self.location = location or get_default_arg("location")
        self.project = project or get_default_arg("project")
        if not project:
            try:
                # E.g. BigQuery(table="project.dataset.table")
                self.project, self.dataset, self.table = table.split(".")
            except (ValueError, AttributeError):
                pass
        try:
            # E.g. BigQuery(table="dataset.table", project="project")
            self.dataset, self.table = self.table.split(".")
        except (ValueError, AttributeError):
            pass
        if self.table and "." in self.table:
            raise ValueError("Table name cannot contain '.'")
        if self.dataset and "." in self.dataset and not project:
            # E.g. BigQuery(dataset="project.dataset")
            self.project, self.dataset = self.dataset.split(".", 1)

        # table_id is the full table ID, e.g. "project.dataset.table"
        self.table_id = None
        if self.table:
            self.table_id = f"{self.dataset}.{self.table}"
        if self.project and self.table_id:
            self.table_id = f"{self.project}.{self.table_id}"
        if self.project and self.dataset:
            self.dataset_id = f"{self.project}.{self.dataset}"

        if not self.project:
            raise ValueError("Project ID not specified.")
        if not self.dataset and self.table:
            raise ValueError("Table name specified without dataset.")

        self.bigquery = ModuleHandler("google.cloud").please_import(
            "bigquery", who_is_calling="BigQuery"
        )
        self.client = ClientHandler(self.bigquery.Client).get(
            project=self.project, location=self.location
        )
        self.exceptions = ModuleHandler("google.api_core.exceptions").please_import(
            who_is_calling="BigQuery"
        )

    def __repr__(self):
        return f"BigQuery({self.table_id})"

    def query(self, sql, job_config=None, schema=None, to_dataframe=False, params=None):
        """
        Executes a query against BigQuery.

        Args:
        - sql (str): SQL query string.
        - job_config (bigquery.QueryJobConfig): Optional configuration for the query job.
        - to_dataframe (bool): If True, returns the results as a pandas DataFrame.
        - params (dict): Optional parameters to pass to the query.

        Returns:
        - Query results as a pandas DataFrame or BigQuery job object.
        """
        if params:
            job_config = self._parse_query_params(params, job_config=job_config)
        output = self.client.query(sql, job_config=job_config)
        if to_dataframe:
            try_import("pandas", "BigQuery.query.to_dataframe")

            output = output.to_dataframe().convert_dtypes()
            if schema and is_dataframe(output):
                pd_schema = Schema(schema).pandas()
                output = output.astype(pd_schema)
        else:
            output = output.result()
            try:
                output = list(output)
            except Exception as e:
                log(f"BigQuery - Error converting query results to list: {e}")
        sql_print = sql
        if params:
            for key, value in params.items():
                sql_print = sql_print.replace(f"@{key}", str(value))
        log(f"BigQuery - Query executed: \n{sql_print}")
        return output

    def read_table(
        self,
        job_config=None,
        to_dataframe=True,
        columns=None,
        filters=None,
        schema=None,
        limit=None,
    ):
        """
        Reads entire table from BigQuery.

        Args:
        - job_config (bigquery.QueryJobConfig): Optional configuration for the query job.
        - to_dataframe (bool): If True, returns the results as a pandas DataFrame.
        - columns (list): Columns to select.
        - filters (list): Filters to apply.
        - schema (list): Schema to apply.
        - limit (int): Limit the number of rows to return.

        Returns:
        - Query results as a pandas DataFrame or BigQuery job object.
        """
        sql, params = (
            SQLBuilder(self.table_id)
            .select(columns)
            .where(filters)
            .limit(limit)
            .build()
        )
        return self.query(
            sql=sql,
            job_config=job_config,
            to_dataframe=to_dataframe,
            schema=schema,
            params=params,
        )

    def read(
        self,
        filepath=None,
        job_config=None,
        to_dataframe=False,
        columns=None,
        filters=None,
        schema=None,
        limit=None,
    ):
        """
        Reads entire table from BigQuery.
        If the Google Storage filepath is provided, it will create an external table and read from it.
        This is generally faster than reading the data from storage using e.g. Pyarrow or Pandas.

        Args:
        - filepath (str): The Google Storage filepath to read from.
        - job_config (bigquery.QueryJobConfig): Optional configuration for the query job.
        - to_dataframe (bool): If True, returns the results as a pandas DataFrame.
        - columns (list): Columns to select.
        - filters (list): Filters to apply.
        - schema (list): Schema to apply.
        - limit (int): Limit the number of rows to return.

        Returns:
        - Query results as a pandas DataFrame or BigQuery job object.
        """
        if filepath is not None and isinstance(filepath, str):
            random_dataset = f"temp_dataset_{os.urandom(8).hex()}"
            random_table = f"temp_table_{os.urandom(8).hex()}"
            table_id = f"{random_dataset}.{random_table}"
            BigQuery(table_id).create_external_table(filepath, schema=schema)
            output = BigQuery(table_id).read_table(
                job_config=job_config,
                to_dataframe=to_dataframe,
                columns=columns,
                filters=filters,
                schema=schema,
                limit=limit,
            )
            BigQuery(dataset=random_dataset).delete()
            return output
        else:
            return self.read_table(
                job_config=job_config,
                to_dataframe=to_dataframe,
                columns=columns,
                filters=filters,
                schema=schema,
                limit=limit,
            )

    def _parse_query_params(self, params, job_config=None):
        """
        Helper to parse query parameters into bigquery.QueryJobConfig.

        Args:
        - params (dict): Query parameters to parse.
        - job_config (bigquery.QueryJobConfig): Optional configuration for the query job.

        Returns:
        - bigquery.QueryJobConfig object: The updated job_config.
        """
        if not job_config:
            job_config = self.bigquery.QueryJobConfig()

        query_params = []

        for key, value in params.items():
            val_type = {"int": "INTEGER", "float": "FLOAT", "str": "STRING"}.get(
                type(value).__name__, "STRING"
            )
            if isinstance(value, list):
                bq_param = self.bigquery.ArrayQueryParameter(key, val_type, value)
            else:
                bq_param = self.bigquery.ScalarQueryParameter(key, val_type, value)
            query_params.append(bq_param)
        job_config.query_parameters = job_config.query_parameters + query_params
        return job_config

    def insert(self, data, schema=None):
        """
        Inserts data into a BigQuery table.

        Args:
        - data (list of dicts | pandas.DataFrame): The data to insert.
        - schema (list of bigquery.SchemaField): Optional schema definition. Required if the table does not exist.

        Returns:
        - True if successful.

        Examples:
        - BigQuery("dataset.table").insert([{"a": 1, "b": "test"}])
        - BigQuery("dataset.table").insert(pd.DataFrame({"a": [1], "b": ["test"]}))
        """
        if is_dataframe(data):
            try_import("pandas_gbq", "BigQuery.insert.dataframe")
            from pandas_gbq import to_gbq

            return to_gbq(
                data,
                destination_table=self.table_id,
                if_exists="append",
                location=self.location,
            )

        if isinstance(data, dict):
            data = orient_dict(data, orientation="index")

        if not schema and isinstance(data, list):
            table = self.client.get_table(self.table_id)  # Make sure the table exists
            errors = self.client.insert_rows_json(table, data)
        elif schema:
            table = self.bigquery.Table(self.table_id, schema=schema)
            errors = self.client.insert_rows_json(table, data)
        else:
            raise ValueError(
                "Data format not supported or schema required for new table."
            )

        if errors == []:
            log(f"BigQuery - Data inserted into: {self.table}")
            return True
        else:
            log(
                f"BigQuery - Errors occurred while inserting data into: {self.table} -- {errors}"
            )
            return False

    def write(self, data, schema=None):
        """
        Writes data to a BigQuery table. If the table does not exist, it will be created. If the table exists, the data will be appended.

        Args:
        - data (list of dicts | pandas.DataFrame): The data to insert.
        - schema (list of bigquery.SchemaField): Optional schema definition. Required if the table does not exist.

        Returns:
        - True if successful.
        """
        if is_dataframe(data):
            self._create_table(data, schema=schema, exists_ok=True, if_exists="append")
            log(f"BigQuery - DataFrame written to {self.table}, schema: {schema}")
            return True

        if isinstance(data, dict):
            data = [data]

        if not schema:
            schema = Schema(data, is_data=True).bigquery()

        try:
            success = self.insert(data, schema=schema)
        except self.exceptions.NotFound:
            success = self.create_table(data, schema=schema, exists_ok=True)
        log(f"BigQuery - Data written to {self.table}, schema: {schema}")
        return success

    def _create_table(
        self,
        data=None,
        schema=None,
        exists_ok=True,
        if_exists=None,
        time_partition_col=None,
        range_partition_col=None,
        cluster_cols=None,
        labels=None,
    ):
        """
        Routine for creating a new BigQuery table.

        Args:
        - data (pandas.DataFrame): The data to insert into the new table.
        - schema (list of bigquery.SchemaField): Schema definition for the new table.
        - exists_ok (bool): If True, the table will be replaced if it already exists.
        - if_exists (str): If "replace", the table will be replaced if it already exists. If "append", the data will be appended to the table.
        - time_partition_col (str): The column to partition the table by. This is used for time-based partitioning.
        - range_partition_col (str): The column to partition the table by. This is used for range-based partitioning.
        - cluster_cols (list): The columns to cluster the table by.
        - labels (dict): Labels to apply to the table.

        Returns:
        - True if successful.
        """
        if time_partition_col and range_partition_col:
            raise ValueError(
                "Cannot have both time and range partitioning in the same table."
            )
        if isinstance(schema, dict):
            schema = dict_to_bigquery_fields(schema)

        if is_dataframe(data):
            try_import("pandas_gbq", "BigQuery.create_table.dataframe")
            from pandas_gbq import to_gbq

            if if_exists is None:
                if_exists = "replace" if exists_ok else "fail"
            return to_gbq(
                data,
                destination_table=self.table_id,
                if_exists=if_exists,
                location=self.location,
            )
        table = self.bigquery.Table(self.table_id, schema=schema)
        if time_partition_col:
            table.time_partitioning = self.bigquery.TimePartitioning(
                field=time_partition_col
            )
        if range_partition_col:
            table.range_partitioning = self.bigquery.RangePartitioning(
                field=range_partition_col
            )
        if cluster_cols:
            table.clustering_fields = cluster_cols
        if labels:
            table.labels = labels

        table = self.client.create_table(table, exists_ok=exists_ok)
        log(f"BigQuery - Table created: {self.table_id}")

        if data is not None:
            self.insert(data, schema=schema)
        return True

    def create_table(self, data=None, schema=None, exists_ok=True):
        """
        Creates a new BigQuery table.

        Args:
        - schema (list of bigquery.SchemaField or dict): The schema definition for the new table.

        Returns:
        - The created bigquery.Table object.

        Examples:
        - schema = [bigquery.SchemaField("name", "STRING"), bigquery.SchemaField("age", "INTEGER")]
        - or schema = {"name": "STRING", "age": "INTEGER"}
        - BigQuery().create_table("new_dataset.new_table", schema)
        """
        try:
            return self._create_table(data=data, schema=schema, exists_ok=exists_ok)
        except self.exceptions.NotFound:
            # Dataset does not exist, so create it and try again
            self.create_dataset(exists_ok=exists_ok)
            return self._create_table(data=data, schema=schema, exists_ok=exists_ok)
        return False

    def _create_external_table(
        self, uri, source_format=None, schema=None, infer_uri=True, exists_ok=True
    ):
        """
        Creates an external table in BigQuery.

        Args:
        - uri (str): The URI of the external data source.
        - source_format (str): The format of the external data source.
        - schema (list of bigquery.SchemaField): The schema of the external table.

        Returns:
        - True if successful.

        Examples:
        - BigQuery("dataset.external_table").create_external_table("gs://bucket/file.parquet")
        """
        if infer_uri:
            uri, source_format, extra_metadata = self._infer_uri(uri, source_format)

        external_config = self.bigquery.ExternalConfig(source_format=source_format)
        external_config.source_uris = [uri]
        # if schema:
        # external_config.schema = schema
        external_config.autodetect = True
        partition_columns = extra_metadata.get("partition_columns", None)
        if source_format == "PARQUET" and partition_columns:
            source_uri_prefix = uri.replace("*", "")
            partition_options = self.bigquery.external_config.HivePartitioningOptions
            options = {
                "mode": "STRINGS",
                "sourceUriPrefix": source_uri_prefix,
            }
            external_config.hive_partitioning = partition_options.from_api_repr(options)

        table = self.bigquery.Table(self.table_id)
        table.external_data_configuration = external_config
        try:
            self.client.create_table(table, exists_ok=exists_ok)
        except self.exceptions.NotFound:
            # Dataset does not exist, so create it and try again
            self.create_dataset()
            self.client.create_table(table)
        log(f"BigQuery - External table created: {self.table_id}")
        return True

    def create_external_table(self, uri, schema=None, exists_ok=True):
        """
        Creates an external table in BigQuery.

        Args:
        - uri (str): The URI of the external data source.
        - source_format (str): The format of the external data source.
        - schema (list of bigquery.SchemaField): The schema of the external table.

        Returns:
        - True if successful.

        Examples:
        - BigQuery("dataset.external_table").create_external_table("gs://bucket/file.parquet")
        """
        try:
            return self._create_external_table(uri, schema=schema, exists_ok=exists_ok)
        except self.exceptions.NotFound:
            # Dataset does not exist, so create it and try again
            self.create_dataset(exists_ok=exists_ok)
            return self._create_external_table(uri, schema=schema, exists_ok=exists_ok)

    def _infer_uri(self, uri, source_format=None):
        """
        Helper method to infer the source format from the URI.

        Args:
        - uri (str): The URI of the external data source, or the path to the file.
        - source_format (str): The format of the external data source.

        Returns:
        - Tuple of URI and source format.
        """
        extra_metadata = {}
        if not uri.startswith("gs://"):
            uri = f"gs://{uri}"

        from gcp_pal.storage import Storage

        if uri.endswith(".parquet"):
            source_format = "PARQUET"
            # Check is the file is partitioned
            from gcp_pal.storage import Storage

            if Storage(uri).isdir():
                from gcp_pal.storage.parquet import _get_partition_cols

                extra_metadata["partition_columns"] = _get_partition_cols(uri)
                if uri.endswith("/"):
                    uri = uri[:-1]
                uri = f"{uri}*"
        elif uri.endswith(".csv"):
            source_format = "CSV"
        elif uri.endswith(".json"):
            source_format = "NEWLINE_DELIMITED_JSON"
        elif uri.endswith(".avro"):
            source_format = "AVRO"
        elif uri.endswith(".orc"):
            source_format = "ORC"
        return uri, source_format, extra_metadata

    def create_dataset(self, exists_ok=True):
        """
        Creates a new BigQuery dataset.

        Returns:
        - True if successful.
        """
        dataset = self.bigquery.Dataset(self.dataset_id)
        dataset = self.client.create_dataset(dataset, exists_ok=exists_ok)
        log(f"BigQuery - Dataset created: {self.dataset_id}")
        return True

    def create(self, data=None, schema=None, exists_ok=True, infer_schema=False):
        """
        Creates a new BigQuery dataset or table.

        Args:
        - data (pandas.DataFrame|str): If a DataFrame, creates a table with the DataFrame schema.
                                       If a string, considers it a URI for an external table.
        - schema (list of bigquery.SchemaField): The schema definition for the new table.
        - exists_ok (bool): If True, the table will be replaced if it already exists.

        Returns:
        - True if successful.
        """
        if not self.table:
            # Working with a dataset.
            return self.create_dataset(exists_ok=exists_ok)

        # Auto infer the schema if not provided.
        if not schema and infer_schema is True:
            schema = Schema(data, is_data=True).bigquery()

        if isinstance(data, str):
            # Working with a table and input is a URI.
            return self.create_external_table(data, schema=schema, exists_ok=exists_ok)
        if isinstance(data, (list, dict)) or is_dataframe(data) or data is None:
            # Working with a table and input is a DataFrame.
            return self.create_table(data=data, schema=schema, exists_ok=exists_ok)
        else:
            raise ValueError("Invalid input for creating a table.")

    def delete_table(self, errors="raise"):
        """
        Deletes a BigQuery table.

        Returns:
        - True if successful.
        """
        try:
            self.client.delete_table(self.table_id)
        except Exception as e:
            log(f"BigQuery - Error deleting table: {e}")
            if errors == "raise":
                raise e
            return False
        log(f"BigQuery - Table deleted: BigQuery/{self.table_id}")
        return True

    def delete_dataset(self, errors="raise"):
        """
        Deletes a BigQuery dataset.

        Returns:
        - True if successful.
        """
        dataset_id = f"{self.project}.{self.dataset}"
        try:
            self.client.delete_dataset(dataset_id, delete_contents=True)
        except Exception as e:
            log(f"BigQuery - Error deleting dataset: {e}")
            if errors == "raise":
                raise e
            return False
        log(f"BigQuery - Dataset deleted: BigQuery/{dataset_id}")
        return True

    def delete(self, errors="ignore"):
        """
        Deletes a BigQuery dataset or table.

        Args:
        - errors (str): If "raise", errors will be raised. If "ignore", errors will be ignored.

        Returns:
        - True if successful.
        """
        if self.table:
            return self.delete_table(errors=errors)
        else:
            return self.delete_dataset(errors=errors)

    def list_datasets(self):
        """
        Lists all datasets in the BigQuery project.

        Returns:
        - List of dataset IDs.
        """
        datasets = list(self.client.list_datasets())
        dataset_ids = [dataset.dataset_id for dataset in datasets]
        log("BigQuery - datasets listed.")
        return dataset_ids

    def list_tables(self, dataset=None):
        """
        Lists all tables in a specified dataset.

        Args:
        - dataset_id (str): The ID of the dataset.

        Returns:
        - List of table IDs within the specified dataset.
        """
        dataset = dataset or self.dataset
        tables = list(self.client.list_tables(dataset))
        table_ids = [table.table_id for table in tables]
        log(f"BigQuery - Tables listed for dataset: {dataset}")
        return table_ids

    def ls(self, dataset=None):
        """
        Lists all datasets in the project, or all tables in a specified dataset.

        Returns:
        - List of dataset IDs or table IDs.
        """
        if self.dataset or dataset:
            return self.list_tables(dataset=dataset)
        else:
            return self.list_datasets()

    def exists(self):
        """
        Checks if a dataset or table exists.

        Returns:
        - True if the dataset or table exists.
        """
        output = False
        if self.table:
            try:
                self.client.get_table(self.table_id)
                output = True
            except self.exceptions.NotFound:
                output = False
        else:
            try:
                self.client.get_dataset(self.dataset_id)
                output = True
            except self.exceptions.NotFound:
                output = False
        return output

    def create_snapshot(self, snapshot_name: str):
        """
        Creates a snapshot of a BigQuery table.

        Args:
        - snapshot_name (str): The name of the snapshot.

        Returns:
        - True if successful.
        """
        success = self.client.create_snapshot(snapshot_name, self.table_id)
        log(f"BigQuery - Snapshot created: {snapshot_name}")
        return success

    def get_table(self):
        """
        Get the BigQuery table object.

        Returns:
        - The bigquery.Table object.
        """
        return self.client.get_table(self.table_id)

    def get_dataset(self):
        """
        Get the BigQuery dataset object.

        Returns:
        - The bigquery.Dataset object.
        """
        return self.client.get_dataset(self.dataset_id)

    def get(self):
        """
        Get the BigQuery dataset or table object.

        Returns:
        - The bigquery.Dataset or bigquery.Table object.
        """
        if self.table:
            return self.get_table()
        else:
            return self.get_dataset()

    def schema(self, as_dict=False):
        """
        Get the schema of the BigQuery table.

        Args:
        - as_dict (bool): If True, returns the schema as a dictionary. The output will be a (nested) dictionary:
                            `{"name": "type", "name": {"name": "type"}}`.

        Returns:
        - tuple[bigquery.SchemaField]: The schema of the table (column definitions).
        """
        schema = self.get_table().schema
        if as_dict:
            schema = bigquery_fields_to_dict(schema)
        return schema

    def set_schema(self, schema):
        """
        Set the schema of the BigQuery table.

        Args:
        - schema (list of bigquery.SchemaField or dict): The schema to set.

        Returns:
        - The updated bigquery.Table object.
        """
        if isinstance(schema, dict):
            schema = dict_to_bigquery_fields(schema)
        table = self.get_table()
        table.schema = schema
        return self.client.update_table(table, ["schema"])


if __name__ == "__main__":
    data = {
        "name": "John",
        "age": 30,
        "city": "New York",
    }
    BigQuery("dataset1.table1").write(data)
    exit()


if __name__ == "__main__":
    import pandas as pd
    from gcp_pal import Storage

    success = {}
    data = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    pa_schema = Schema(data, is_data=True).pyarrow()
    bq_schema = Schema(data, is_data=True).bigquery()
    bucket_name = f"test_bucket_vita_123"
    file_name = f"gs://{bucket_name}/file.csv"

    Storage(bucket_name).create()

    data.to_csv(file_name, index=False)
    success[0] = Storage(file_name).exists()
    success[1] = not Storage(file_name).isdir()
    success[2] = Storage(file_name).isfile()

    dataset_name = f"test_dataset_123"
    table_name = f"test_ext_table123"
    table_id = f"{dataset_name}.{table_name}"
    BigQuery(table_id).create(file_name, schema=bq_schema)
    success[3] = BigQuery(table_id).exists()

    queried_df = BigQuery().read(file_name, schema=bq_schema, to_dataframe=True)

    queried_df = queried_df.sort_values("a").reset_index(drop=True)
    queried_df = queried_df[data.columns]

    print(queried_df)
    print(data)
    print(queried_df.dtypes)
    print(data.dtypes)
    exit()


if __name__ == "__main__":
    import pandas as pd
    from uuid import uuid4
    from gcp_pal.storage import Storage, Parquet

    success = {}
    data = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    pa_schema = Schema(data, is_data=True).pyarrow()
    bq_schema = Schema(data, is_data=True).bigquery()
    bucket_name = f"test_bucket_{uuid4().hex}"
    file_name = f"gs://{bucket_name}/file.parquet"
    Storage(bucket_name).create()

    Parquet(file_name).write(data, partition_cols=["a"], schema=pa_schema)

    dataset_name = f"test_dataset_{uuid4().hex}"
    table_name = f"test_ext_table"
    table_id = f"{dataset_name}.{table_name}"
    BigQuery(table_id).create(file_name, schema=bq_schema)

    queried_df = BigQuery().read(file_name, schema=bq_schema)

    print(queried_df)
    print(data)
    exit()


if __name__ == "__main__":
    # Example usage
    from google.cloud import bigquery

    dataset = "clean"
    print(BigQuery().ls())
    print(BigQuery(dataset=dataset).get_dataset().dataset_id)
    BigQuery(f"{dataset}.new_table1").create_table(
        schema=[
            bigquery.SchemaField("name", "STRING"),
            bigquery.SchemaField("age", "INTEGER"),
        ]
    )
    BigQuery(f"{dataset}.new_table1").insert(
        [{"name": "John", "age": 30}, {"name": "Jane", "age": 25}]
    )
    df = BigQuery(f"{dataset}.new_table1").read(
        columns=["name", "age"],
        filters=[("age", ">", 25), ("name", "like", "J%")],
        to_dataframe=True,
    )

    print(df)

    print(BigQuery().ls())  # List all datasets
    print(BigQuery(dataset=f"{dataset}").ls())  # List all tables in the "clean" dataset
    BigQuery(dataset=f"{dataset}").delete()



import json
import random

from gcp_pal.pydocker import Docker
from gcp_pal.utils import log, ClientHandler, ModuleHandler, get_default_arg


class CloudRun:

    def __init__(
        self,
        name=None,
        project=None,
        location=None,
        job=False,
        service_account=None,
    ):
        self.job = job
        self.type = "job" if job else "service"
        self.service_account = service_account
        self.project = project or get_default_arg("project")
        self.location = location or get_default_arg("location")
        self.parent = f"projects/{self.project}/locations/{self.location}"
        if isinstance(name, str) and name.startswith("projects/"):
            name = name.split("/")[-1]
        self.name = name
        self.full_name = f"{self.parent}/{self.type}s/{self.name}"
        self.image_url = None

        self.run = ModuleHandler("google.cloud").please_import(
            "run_v2", who_is_calling="CloudRun"
        )
        self.types = self.run.types
        self.client = ClientHandler(self.run.ServicesClient).get()
        self.jobs_client = ClientHandler(self.run.JobsClient).get()
        self.exceptions = ModuleHandler("google.api_core.exceptions").please_import(
            who_is_calling="CloudRun"
        )

    def ls_jobs(self, active_only=False, full_id=False):
        """
        List all jobs in the project.
        """
        jobs = self.jobs_client.list_jobs(parent=self.parent)
        if active_only:
            jobs = [f for f in jobs if f.terminal_condition.type_ == "Ready"]
        output = [x.name for x in jobs]
        if not full_id:
            output = [f.split("/")[-1] for f in output]
        return output

    def ls_services(self, active_only=False, full_id=False):
        """
        List all services in the project.
        """
        services = self.client.list_services(parent=self.parent)
        if active_only:
            services = [
                f for f in services if all([x.percent > 0 for x in f.traffic_statuses])
            ]
        output = [x.name for x in services]
        if not full_id:
            output = [f.split("/")[-1] for f in output]
        return output

    def ls(self, active_only=False, full_id=False):
        """
        List all services or jobs in the project.

        Args:
        - active_only (bool): Only list active services or jobs. Default is False.
        - full_id (bool): Return full resource IDs, e.g. 'projects/my-project/locations/europe-west2/services/my-service'. Default is False.

        Returns:
        - (list): A list of service or job names.
        """
        if self.job:
            return self.ls_jobs(active_only=active_only, full_id=full_id)
        return self.ls_services(active_only=active_only, full_id=full_id)

    def get(self):
        """
        Get a service or job by name.

        Returns:
        - (Service) or (Job): The service or job object.
        """
        if self.job:
            return self.jobs_client.get_job(name=self.full_name)
        return self.client.get_service(name=self.full_name)

    def status(self):
        """
        Get the status of a service or job.

        Returns:
        - (str): The status of the service or job. Either 'Active' or 'Inactive'.
        """
        if self.job:
            status = self.get().terminal_condition.type_
            output = "Active" if status == "Ready" else "Inactive"
            log(f"Cloud Run - Job '{self.name}' is {output}.")
        else:
            status = self.get().traffic[0].percent > 0
            output = "Active" if status else "Inactive"
            log(f"Cloud Run - Service '{self.name}' is {output}.")
        return output

    def state(self):
        """
        Alias for `status`.
        """
        return self.status()

    def exists(self):
        """
        Check if a service or job exists.

        Returns:
        - (bool): True if the service or job exists, False otherwise.
        """
        try:
            self.get()
            return True
        except:
            return False

    def build_and_push_docker_image(
        self, path=".", image_tag="latest", dockerfile="Dockerfile"
    ):
        """
        Build and push a Docker image to GCR.

        Args:
        - path (str): The path to the build context.
        - image_tag (str): The tag to apply to the image.
        - dockerfile (str): The path to the Dockerfile from the context.

        Returns:
        - (str): The GCR location of the image (e.g. 'gcr.io/my-project/my-image:latest')
        """
        docker = Docker(name=self.name, project=self.project, tag=image_tag)
        gcr_location = docker.build_and_push(path=path, dockerfile=dockerfile)
        self.image_url = gcr_location
        if gcr_location is None:
            raise ValueError("Failed to build and push Docker image.")
        return gcr_location

    def deploy_service(
        self,
        image_url=None,
        env_vars_file=None,
        memory="512Mi",
        min_instances=0,
        max_instances=0,
        container_kwargs={},
        service_kwargs={},
        wait_to_complete=True,
    ):
        """
        Deploy a Cloud Run service using a Docker image.
        Args:
        - env_vars_file (str): YAML file path for environment variables.
        - memory (str): Amount of memory to allocate to each service instance.
                        Also accepts integer vaues in MB. Default is '512Mi'.
        - min_instances (int): Minimum number of instances.
        - max_instances (int): Maximum number of instances. Set to 0 for auto-scaling.
        - container_kwargs (dict): Additional container configuration.
        - service_kwargs (dict): Additional service configuration.
        """
        if isinstance(memory, int):
            memory = f"{memory}Mi"
        image_url = image_url or self.image_url
        # gcloud run deploy seems to set HOST=0.0.0.0 by default?
        default_env_vars = {"HOST": "0.0.0.0"}
        service = self.run.Service(
            template={
                "containers": [
                    {
                        "image": image_url,
                        "env": self._parse_env_vars(
                            env_vars_file, default=default_env_vars
                        ),
                        "resources": self.types.ResourceRequirements(
                            limits={"memory": memory}
                        ),
                        **container_kwargs,
                    }
                ],
                "scaling": {
                    "min_instance_count": min_instances,
                    "max_instance_count": max_instances,
                },
            },
            **service_kwargs,
        )
        service_exists = self.exists()
        if not service_exists:
            log(f"Cloud Run - Creating service '{self.name}'...")
            args = {"parent": self.parent, "service": service, "service_id": self.name}
            request = self.types.CreateServiceRequest(**args)
            response = self.client.create_service(request=request)
        else:
            log(f"Cloud Run - Updating service '{self.name}'...")
            # Add 'name' to service.template.containers
            service.name = self.full_name
            request = self.types.UpdateServiceRequest(service=service)
            response = self.client.update_service(request=request)
        if wait_to_complete:
            log(f"Cloud Run - Waiting for service '{self.name}' to complete...")
            response.result(timeout=300)
        return response

    def deploy_job(
        self,
        image_url=None,
        env_vars_file=None,
        memory="512Mi",
        container_kwargs={},
        job_kwargs={},
    ):
        """
        Deploy a Cloud Run job using a Docker image.
        Args:
        - env_vars_file (str): YAML file path for environment variables.
        - memory (str): Amount of memory to allocate to the job.
        - container_kwargs (dict): Additional container configuration.
        - job_kwargs (dict): Additional job configuration.
        """
        image_url = image_url or self.image_url
        job = self.run.Job(
            name=self.full_name,
            template={
                "containers": [
                    {
                        "image": image_url,
                        "env": self._parse_env_vars(env_vars_file),
                        "resources": self.types.ResourceRequirements(
                            limits={"memory": memory}
                        ),
                        **container_kwargs,
                    }
                ]
            },
            **job_kwargs,
        )
        response = self.jobs_client.create_job(parent=self.parent, job=job)
        return response

    def uri(self):
        """
        Get the URI of the service.
        """
        return self.get().uri

    def call(
        self,
        data={},
        errors="ignore",
        to_json=True,
        service_account=None,
    ):
        """
        Calls a cloud function.

        Args:
        - data (dict|str): The data to send to the cloud function. If a dict, it will be converted to JSON. Defaults to {}.
        - errors (str): How to handle errors. Options are "ignore", "raise" or "log". Defaults to "ignore".
        - to_json (bool): Whether to convert the response to JSON. Defaults to True.
        - service_account (str): The service account to use for the call.

        Returns:
        - (dict) The response from the cloud function.
        """
        if isinstance(data, dict):
            data = json.dumps(data)

        from gcp_pal import Request

        uri = self.uri()
        service_account = service_account or self.service_account
        output = Request(uri, service_account=service_account).post(data)
        if output.status_code != 200:
            msg = f"Cloud Function - Error calling '{self.name}': {output.text}"
            if errors == "raise":
                raise Exception(msg)
            elif errors == "log":
                log(msg)
        if to_json:
            try:
                output = output.json()
            except json.JSONDecodeError:
                pass
        return output

    def invoke(self, **kwargs):
        """
        Alias for call method.
        """
        return self.call(**kwargs)

    def _parse_env_vars(self, yaml_file, default={}):
        """
        Parse YAML file to extract environment variables as a list of EnvVar objects.
        """
        from gcp_pal.utils import load_yaml

        if yaml_file is None:
            return []
        data = load_yaml(yaml_file)
        data = {**default, **data}
        return [self.types.EnvVar(name=k, value=str(v)) for k, v in data.items()]

    def deploy(
        self,
        path=".",
        image_tag="random",
        dockerfile="Dockerfile",
        **kwargs,
    ):
        """
        Deploy a Cloud Run service or job.

        Args:
        - path (str): The path to the build context or the image URL.
        - image_tag (str): The tag to apply to the image. Default is 'random', which will generate a unique tag.
        - dockerfile (str): The path to the Dockerfile from the context.
        - kwargs: Additional arguments to pass to the deploy_service or deploy_job method.

        Returns:
        - (Service) or (Job): The service or job object.
        """
        if image_tag == "random":
            random_tag = random.getrandbits(64)
            image_tag = f"{random_tag:x}"
        if (
            path.startswith("gs:")
            or path.startswith("http")
            or path.startswith("gcr.io/")
        ):
            log(f"Cloud Run - Deploying {path} directly.")
            image_url = path
        else:
            log(f"Cloud Run - Building and pushing Docker image from {path}.")
            image_url = self.build_and_push_docker_image(
                path=path, image_tag=image_tag, dockerfile=dockerfile
            )
        if self.job:
            return self.deploy_job(image_url, **kwargs)
        else:
            return self.deploy_service(image_url, **kwargs)

    def create(
        self,
        path=".",
        image_tag="random",
        dockerfile="Dockerfile",
        **kwargs,
    ):
        """
        (Alias of `deploy`) Deploy a Cloud Run service or job.

        Args:
        - path (str): The path to the build context or the image URL.
        - image_tag (str): The tag to apply to the image. Default is 'random', which will generate a unique tag.
        - dockerfile (str): The path to the Dockerfile from the context.
        - kwargs: Additional arguments to pass to the deploy_service or deploy_job method.

        Returns:
        - (Service) or (Job): The service or job object.
        """
        return self.deploy(path, image_tag, dockerfile, **kwargs)

    def delete(self, wait_to_complete=True, errors="ignore"):
        """
        Delete a service or job.

        Returns:
        - None
        """
        try:
            if self.job:
                result = self.jobs_client.delete_job(name=self.full_name)
                if wait_to_complete:
                    result.result()
                log(f"Cloud Run - Deleted job '{self.name}'.")
            else:
                result = self.client.delete_service(name=self.full_name)
                if wait_to_complete:
                    result.result()
                log(f"Cloud Run - Deleted service '{self.name}'.")
        except self.exceptions.NotFound:
            if errors == "ignore":
                log(f"Cloud Run - Service '{self.name}' not found to delete.")
                return None
            else:
                raise


if __name__ == "__main__":
    CloudRun("test-app").deploy(path="samples/cloud_run")
    output = CloudRun("test-app").call(data={"data": 16})
    print(output)
    CloudRun("test-app").delete()
